from typing import Dict

from ..protocols import Protocol

def revert_mappings(protocol: Protocol, test_predictions: Dict, class_int2str: Dict[int, str]) -> Dict[str, str]:
    # If residue-to-class problem, map the integers back to the class labels (single letters)
    if protocol == Protocol.residue_to_class:
        return {seq_id: "".join([class_int2str[p] for p in prediction]
                                if type(prediction) == list else class_int2str[prediction])
                for seq_id, prediction in test_predictions.items()}

    # If sequence/residues-to-class problem, map the integers back to the class labels (whatever length)
    elif protocol == Protocol.sequence_to_class or protocol == Protocol.residues_to_class:
        return {seq_id: class_int2str[prediction] for seq_id, prediction in test_predictions.items()}
    else:
        return test_predictions



from typing import List, Dict, Tuple

from .biotrainer_seq_record import BiotrainerSequenceRecord

from ..utilities.constants import INTERACTION_INDICATOR


def merge_protein_interactions(seq_records: List[BiotrainerSequenceRecord]) -> Dict[str, Dict[str, str]]:
    """
    :param seq_records: a list of BiotrainerSequenceRecord
    :return: A dictionary of ids and their attributes
    """
    result = {}

    seq_id_to_hash = {seq_record.seq_id: seq_record.get_hash() for seq_record in seq_records}

    for seq_record in seq_records:
        seq_id = seq_record.seq_id
        seq_hash = seq_record.get_hash()
        interactor = seq_record.get_ppi()
        interactor_hash = seq_id_to_hash[interactor]
        if not interactor:
            raise ValueError(f"Sequence {seq_id} does not have a valid interactor!")

        interaction_id = f"{seq_hash}{INTERACTION_INDICATOR}{interactor_hash}"
        interaction_id_flipped = f"{interactor_hash}{INTERACTION_INDICATOR}{seq_hash}"

        # Check that target annotations and sets are consistent:
        for int_id in [interaction_id, interaction_id_flipped]:
            if int_id in result.keys():
                if seq_record.get_target() != result[int_id]["TARGET"]:
                    raise ValueError(f"Interaction multiple times present in fasta file, but TARGET=values are "
                                     f"different for {int_id}!")
                if seq_record.get_set() != result[int_id]["SET"]:
                    raise ValueError(f"Interaction multiple times present in fasta file, but SET=sets are "
                                     f"different for {int_id}!")

        if interaction_id_flipped not in result.keys():
            result[interaction_id] = seq_record.attributes

    return result


def get_split_lists(id2sets: dict) -> Tuple[List[str], List[str], Dict[str, List[str]], List[str]]:
    """ Parse splits from input_file and check that all set annotations are given and correct """
    training_ids = []
    validation_ids = []
    testing_ids = {}
    prediction_ids = []

    incorrect_sets_exception = lambda idx, split: ValueError(f"FASTA header must contain SET. "
                                                             f"SET must be either 'train', "
                                                             f"'val', 'test'/'test{{nr}}' or 'pred'. "
                                                             f"Id: {idx}; SET={split}")

    for idx, split in id2sets.items():
        split = split.lower() if split else ""
        if split == "" or (split.lower() not in ["train", "val", "pred"] and "test" not in split.lower()):
            raise incorrect_sets_exception(idx, split)
        match split:
            case "train":
                training_ids.append(idx)
            case "val":
                validation_ids.append(idx)
            case "pred":
                prediction_ids.append(idx)
            case _:
                if "test" in split:
                    if split not in testing_ids:
                        testing_ids[split] = []
                    testing_ids[split].append(idx)
                else:
                    raise incorrect_sets_exception(idx, split)

    return training_ids, validation_ids, testing_ids, prediction_ids



import torch
import torch.nn as nn

from .biotrainer_model import BiotrainerModel


# Feed-Forward Neural Network (FNN) with two linear layers connected by a non-lin
class FNN(BiotrainerModel):
    def __init__(
            self, n_classes: int, n_features: int,
            bottleneck_dim: int = 32, dropout_rate: float = 0.25,
            **kwargs
    ):
        super(FNN, self).__init__()

        self.classifier = nn.Sequential(
            nn.Linear(n_features, bottleneck_dim),  # n_features x 32
            nn.ReLU(),
            nn.Dropout(dropout_rate),
            nn.Linear(bottleneck_dim, n_classes)
        )

    def forward(self, x: torch.Tensor, *args, **kwargs) -> torch.Tensor:
        """
            L = protein length
            B = batch-size
            F = number of features (1024 for embeddings)
            N = number of classes (e.g. 9 for conservation)
        """
        # IN: X = (B x L x F)
        Yhat = self.classifier(x)  # OUT: Yhat_consurf = (B x L x N)
        return Yhat


class DeeperFNN(BiotrainerModel):
    def __init__(
            self, n_classes: int, n_features: int,
            bottleneck_dim: int = 128, dropout_rate: float = 0.25,
            **kwargs
    ):
        super(DeeperFNN, self).__init__()

        self.classifier = nn.Sequential(
            nn.Linear(n_features, bottleneck_dim),  # n_features x 128 (for default)
            nn.ReLU(),
            nn.Dropout(dropout_rate),
            nn.Linear(bottleneck_dim, int(bottleneck_dim / 2)),  # 128 x 64 (for default)
            nn.ReLU(),
            nn.Dropout(dropout_rate),
            nn.Linear(int(bottleneck_dim / 2), int(bottleneck_dim / 4)),  # 64 x 32 (for default)
            nn.ReLU(),
            nn.Dropout(dropout_rate),
            nn.Linear(int(bottleneck_dim / 4), n_classes)  # 32 x n_classes (for default)
        )

    def forward(self, x: torch.Tensor, *args, **kwargs) -> torch.Tensor:
        """
            L = protein length
            B = batch-size
            F = number of features (1024 for embeddings)
            N = number of classes (e.g. 9 for conservation)
        """
        # IN: X = (B x L x F)
        Yhat = self.classifier(x)  # OUT: Yhat_consurf = (B x L x N)
        return Yhat



import torch

from typing import Tuple


def get_mean_and_confidence_bounds(values: torch.Tensor, dimension: int, confidence_level: float) -> \
        Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """
    Calculates the mean and confidence range for the given values. Used for bootstrapping error reporting and
    monte carlo dropout.

    :param values: Predicted values
    :param dimension: Dimension to consider for values tensor
    :param confidence_level: Confidence level for result confidence intervals (0.05 => 95% percentile)
    :return: Tuple: Tensor with mean over values and confidence range for each value
    """
    if not 0 < confidence_level < 1:
        raise ValueError(f"Confidence level must be between 0 and 1, given: {confidence_level}!")

    values_float = values.float()

    mean = torch.mean(values_float, dim=dimension)

    # Calculate percentiles from actual distribution
    lower_percentile = (confidence_level / 2) * 100
    upper_percentile = (1 - confidence_level / 2) * 100

    lower_bound = torch.quantile(values_float, lower_percentile / 100, dim=dimension)
    upper_bound = torch.quantile(values_float, upper_percentile / 100, dim=dimension)

    return mean, lower_bound, upper_bound



from .seeder import seed_all
from .version import __version__
from .revert_mappings import revert_mappings
from .logging import get_logger, setup_logging
from .execution_environment import is_running_in_notebook
from .hashing import calculate_model_hash, calculate_sequence_hash
from .data_classes import Split, SplitResult, EmbeddingDatasetSample, SequenceDatasetSample, EpochMetrics
from .cuda_device import get_device, is_device_cpu, is_device_cuda, is_device_mps, get_device_memory
from .constants import (
    SEQUENCE_PAD_VALUE,
    MASK_AND_LABELS_PAD_VALUE,
    INTERACTION_INDICATOR,
    METRICS_WITHOUT_REVERSED_SORTING,
    RESIDUE_TO_VALUE_TARGET_DELIMITER,
    AMINO_ACIDS,
)


__all__ = [
    'seed_all',
    'get_logger',
    'calculate_model_hash',
    'calculate_sequence_hash',
    'setup_logging',
    'get_device',
    'is_device_cpu',
    'is_device_cuda',
    'is_device_mps',
    'get_device_memory',
    'is_running_in_notebook',
    'SEQUENCE_PAD_VALUE',
    'MASK_AND_LABELS_PAD_VALUE',
    'INTERACTION_INDICATOR',
    'RESIDUE_TO_VALUE_TARGET_DELIMITER',
    'METRICS_WITHOUT_REVERSED_SORTING',
    'AMINO_ACIDS',
    'Split',
    'SplitResult',
    'EmbeddingDatasetSample',
    'SequenceDatasetSample',
    'EpochMetrics',
    'revert_mappings',
    '__version__'
]



import torch

from pathlib import Path
from typing import Union, List, Tuple

from .config_option import ConfigOption, ConfigConstraints, ConfigKey

from ..protocols import Protocol
from ..utilities import get_device


def general_config(protocol: Protocol) -> Tuple[ConfigKey, List[ConfigOption]]:
    general_category = "general"
    return ConfigKey.ROOT, [
        ConfigOption(name="device",
                     description="Select the device on which computations will be performed.",
                     category=general_category,
                     required=False,
                     default=str(get_device().type),
                     constraints=ConfigConstraints(
                         type=str,
                         allowed_values=["cpu", "mps"] + [f"cuda:{i}" for i in
                                                          range(torch.cuda.device_count())] + ["cuda"]
                         if torch.cuda.is_available() else [])),
        ConfigOption(name="interaction",
                     description="Select a method of protein-protein interaction, "
                                 "either by pair-wise multiplying or concatenating embeddings. "
                                 "Enabling it will interpret the given sequence file as an interaction dataset.",
                     category=general_category,
                     required=False,
                     constraints=ConfigConstraints(
                         type=str,
                         allowed_values=["multiply", "concat"],
                         allowed_protocols=Protocol.using_per_sequence_embeddings())),
        ConfigOption(name="seed",
                     description="Specify a seed value for all random number generators to ensure reproducibility.",
                     category=general_category,
                     required=False,
                     default=42,
                     constraints=ConfigConstraints(
                         type=int,
                     ),
                     ),
        ConfigOption(name="save_split_ids",
                     description="Choose whether to save the identifiers of data splits "
                                 "(i.e. training, validation, test) in the result file.",
                     category=general_category,
                     required=False,
                     default=False,
                     constraints=ConfigConstraints(
                         type=bool,
                     ),
                     ),
        ConfigOption(name="sanity_check",
                     description="Enable or disable sanity checks at the end of training to get immediate feedback "
                                 "about the quality of the trained model.",
                     category=general_category,
                     required=False,
                     default=True,
                     constraints=ConfigConstraints(
                         type=bool,
                     ),
                     ),
        ConfigOption(name="ignore_file_inconsistencies",
                     description="Choose whether to ignore inconsistencies in file-related configurations, "
                                 "such as labels that are not present in the sequence file.",
                     category=general_category,
                     required=False,
                     default=False,
                     constraints=ConfigConstraints(
                         type=bool,
                     ),
                     ),
        ConfigOption(name="external_writer",
                     description="Use an external writer such as tensorboard to track training progress.",
                     category=general_category,
                     required=False,
                     default="tensorboard",
                     constraints=ConfigConstraints(
                         type=str,
                         allowed_values=["tensorboard", "none"],
                     ),
                     ),
        ConfigOption(name="output_dir",
                     description="Define the directory where output files and results will be stored.",
                     category=general_category,
                     required=False,
                     default="output",
                     constraints=ConfigConstraints(
                         type=Union[str, Path],
                     ),
                     ),
        ConfigOption(name="bootstrapping_iterations",
                     description="Specify the number of bootstrap iterations for the trained model on the test set. "
                                 "Using a value of 0 disables bootstrapping.",
                     category=general_category,
                     required=False,
                     default=30,
                     constraints=ConfigConstraints(
                         type=int,
                         gte=0
                     ),
                     ),
        ConfigOption(name="validate_input",
                     description="Run checks on the given input_file data before training.",
                     category=general_category,
                     required=False,
                     default=True,
                     constraints=ConfigConstraints(
                         type=bool,
                     ),
                     )
    ]



import torch
import numpy as np

from ..interfaces import EmbedderInterface


class RandomEmbedder(EmbedderInterface):
    """
    Baseline embedder: Generate random 128xL embedding vectors.

    This embedder is meant to be used as a naive baseline to compare against other pretrained embedders.
    """

    embedding_dimension = 128
    name = "random_embedder"

    def __init__(self):
        self.rng = np.random.default_rng()

    def _embed_single(self, sequence: str) -> torch.tensor:
        return torch.tensor(self.rng.random((len(sequence), self.embedding_dimension), dtype=np.float32))

    @staticmethod
    def reduce_per_protein(embedding: torch.tensor) -> torch.tensor:
        """Reduce via mean"""
        return embedding.mean(axis=0)



from __future__ import annotations

import os
import torch
import onnx
import numpy as np

from pathlib import Path
from torch.utils.data import DataLoader
from typing import Union, Optional, Dict, Iterable, Tuple, Any, List

from ..losses import get_loss
from ..models import get_model
from ..protocols import Protocol
from ..optimizers import get_optimizer
from ..output_files import InferenceOutputManager
from ..datasets import get_dataset, get_embeddings_collate_function
from ..solvers import get_solver, get_mean_and_confidence_bounds, MetricsCalculator
from ..utilities import seed_all, EmbeddingDatasetSample, MASK_AND_LABELS_PAD_VALUE, revert_mappings


class Inferencer:

    def __init__(
            self,
            iom: InferenceOutputManager,
    ):
        self.protocol = iom.protocol()
        self.embedder_name = iom.embedder_name()
        self.embedding_dimension = iom.n_features()
        self.class_int2str = iom.class_int2str()
        self.class_str2int = iom.class_str2int()
        self.device = iom.device()
        self.collate_function = get_embeddings_collate_function(self.protocol)

        self.solvers_and_loaders_by_split = self._create_solvers_and_loaders_by_split(iom)
        print(f"Got {len(self.solvers_and_loaders_by_split.keys())} split(s): "
              f"{', '.join(self.solvers_and_loaders_by_split.keys())}")

    @classmethod
    def create_from_out_file(cls, out_file_path: str,
                             automatic_path_correction: bool = True,
                             ) -> Tuple[Inferencer, InferenceOutputManager]:
        """
        Create the inferencer object from the out.yml file generated by biotrainer.
        Reads the out.yml file without the split ids which would blow up the file unnecessarily.

        :param out_file_path: Path to out.yml file generated by biotrainer
        :param automatic_path_correction: If True, the method tries to correct the log_dir path if it is not found.
                                          In this case, checkpoints are searched at
                                          out_file_path/model_choice/embedder_name

        :return: Tuple with Inferencer object configured with the output variables from the out.yml file,
                 and the output variables as dict
        """
        inference_output_manager = InferenceOutputManager(Path(out_file_path),
                                                          automatic_path_correction=automatic_path_correction)
        return cls(inference_output_manager), inference_output_manager

    def _create_solvers_and_loaders_by_split(self, iom: InferenceOutputManager) -> Dict[str, Tuple[Any, Any]]:
        result_dict = {}
        splits = iom.training_results().keys()
        log_dir = iom.log_dir()
        split_checkpoints = {file.split("_checkpoint.")[0]: file for file in os.listdir(log_dir) if
                             (Path(log_dir) / Path(file)).is_file() and file.endswith(".safetensors")}
        for split_name in splits:
            # Ignore average or best result
            if "average" in split_name or "best" in split_name:
                continue
            split_config = iom.split_config(split_name=split_name)
            split_config["protocol"] = self.protocol
            split_config["device"] = self.device
            # Positional arguments
            #model_choice = split_config.pop("model_choice")
            #n_classes = split_config.pop("n_classes")
            #loss_choice = split_config.pop("loss_choice")
            #optimizer_choice = split_config.pop("optimizer_choice")
            #learning_rate = split_config.pop("learning_rate")
            #log_dir = split_config.pop("log_dir")
            #disable_pytorch_compile = split_config.pop("disable_pytorch_compile")
            checkpoint_path = Path(log_dir) / Path(split_checkpoints[split_name])

            model = get_model(**split_config)
            loss_function = get_loss(**split_config)
            optimizer = get_optimizer(model_parameters=model.parameters(),
                                      **split_config
                                      )

            solver = get_solver(name=split_name, network=model, optimizer=optimizer,
                                loss_function=loss_function,
                                **split_config)
            solver.load_checkpoint(checkpoint_path=checkpoint_path, resume_training=False,
                                   disable_pytorch_compile=iom.disable_pytorch_compile())

            def dataloader_function(dataset):
                return DataLoader(dataset=dataset, batch_size=split_config["batch_size"],
                                  shuffle=False, drop_last=False,
                                  collate_fn=self.collate_function)

            result_dict[split_name] = (solver, dataloader_function)
        return result_dict

    def _convert_class_str2int(self, to_convert: str):
        if type(to_convert) is str:
            if self.protocol in Protocol.per_residue_protocols():
                return [self.class_str2int[t] for t in to_convert]
            else:
                return self.class_str2int[to_convert]
        else:
            return to_convert

    @staticmethod
    def _pad_tensor(protocol: Protocol, target: Union[Any, torch.Tensor], length_to_pad: int, device):
        target_tensor = torch.as_tensor(target, device=device)
        if protocol in Protocol.per_residue_protocols():
            if target_tensor.shape[0] < length_to_pad:
                padding_size = length_to_pad - target_tensor.shape[0]
                padding = torch.full((padding_size,), MASK_AND_LABELS_PAD_VALUE, dtype=target_tensor.dtype,
                                     device=device)
                return torch.cat([target_tensor, padding])
            else:
                return target_tensor
        else:
            return target_tensor

    def _convert_target_dict(self, target_dict: Dict[str, str]):
        if self.protocol in Protocol.classification_protocols():
            if self.protocol in Protocol.per_residue_protocols():
                max_prediction_length = len(max(target_dict.values(), key=len))
                return {seq_id: self._pad_tensor(protocol=self.protocol, target=self._convert_class_str2int(prediction),
                                                 length_to_pad=max_prediction_length, device=self.device)
                        for seq_id, prediction in target_dict.items()}
            else:
                return {seq_id: torch.tensor(self._convert_class_str2int(prediction),
                                             device=self.device)
                        for seq_id, prediction in target_dict.items()}
        return {seq_id: torch.tensor(prediction,
                                     device=self.device)
                for seq_id, prediction in target_dict.items()}

    def _load_solver_and_dataloader(self, embeddings: Union[Iterable, Dict],
                                    split_name, targets: Optional[List] = None):
        if split_name not in self.solvers_and_loaders_by_split.keys():
            raise ValueError(f"Unknown split_name {split_name} for given configuration!")

        if isinstance(embeddings, Dict):
            embeddings_dict = embeddings
        else:
            embeddings_dict = {str(idx): embedding for idx, embedding in enumerate(embeddings)}

        if targets and self.protocol in Protocol.classification_protocols():
            targets = [self._convert_class_str2int(target) for target in targets]

        solver, loader = self.solvers_and_loaders_by_split[split_name]
        dataset = get_dataset(samples=[
            EmbeddingDatasetSample(seq_id, torch.tensor(np.array(embedding)),
                                   torch.empty(1) if not targets else torch.tensor(np.array(targets[idx])))
            for idx, (seq_id, embedding) in enumerate(embeddings_dict.items())
        ], finetuning=False, random_masking=False, mask_token=None, class_str2int=None)
        dataloader = loader(dataset)
        return solver, dataloader

    def convert_all_checkpoints_to_safetensors(self) -> None:
        """
        Converts all checkpoint files for the splits from .pt to .safetensors, if not already stored as .safetensors
        """
        for split_name, (solver, _) in self.solvers_and_loaders_by_split.items():
            if "pt" in solver.checkpoint_type:
                solver.save_checkpoint(solver.start_epoch)

    def convert_to_onnx(self, output_dir: Optional[str] = None) -> List[str]:
        """
        Converts the model to ONNX format for the given embedding dimension.

        :param output_dir: The directory to save the ONNX files. If not provided, the ONNX files will be saved in the
            solver's log directory. Defaults to None.

        :return: A list of file paths where the ONNX files are saved.
        """
        result_file_paths = []
        for split_name, (solver, _) in self.solvers_and_loaders_by_split.items():
            onnx_save_path = solver.save_as_onnx(embedding_dimension=self.embedding_dimension, output_dir=output_dir)
            result_file_paths.append(onnx_save_path)
        return result_file_paths

    def from_embeddings(self, embeddings: Union[Iterable, Dict], targets: Optional[List] = None,
                        split_name: str = "hold_out",
                        include_probabilities: bool = False) -> Dict[str, Union[Dict, str, int, float]]:
        """
        Calculate predictions from embeddings.

        :param embeddings: Iterable or dictionary containing the input embeddings to predict on.
        :param targets: Iterable that contains the targets to calculate metrics
        :param split_name: Name of the split to use for prediction. Default is "hold_out".
        :param include_probabilities: If True, the probabilities used to predict classes are also reported.
                                      Is only useful for classification tasks, otherwise the "probabilities" are the
                                      same as the predictions.
        :return: Dictionary containing the following sub-dictionaries:
                 - 'metrics': Calculated metrics if 'targets' are given, otherwise 'None'.
                 - 'mapped_predictions': Class or value prediction from the given embeddings.
                 - 'mapped_probabilities': Probabilities for classification tasks if include_probabilities is True.
                 Predictions and probabilities are either 'mapped' to keys from an embeddings dict or indexes if
                 embeddings are given as a list.
        """
        solver, dataloader = self._load_solver_and_dataloader(embeddings, split_name, targets)

        inference_dict = solver.inference(dataloader, calculate_test_metrics=targets is not None)
        predictions = inference_dict["mapped_predictions"]

        # For class predictions, revert from int (model output) to str (class name)
        inference_dict["mapped_predictions"] = revert_mappings(protocol=self.protocol, test_predictions=predictions,
                                                               class_int2str=self.class_int2str)
        inference_dict["mapped_probabilities"] = {k: v.cpu().tolist() if v is torch.tensor else v for k, v in
                                                  inference_dict["mapped_probabilities"].items()}

        if not include_probabilities:
            return {k: v for k, v in inference_dict.items() if k != "mapped_probabilities"}
        else:
            return inference_dict

    def from_embeddings_with_bootstrapping(self, embeddings: Union[Iterable, Dict], targets: List,
                                           split_name: str = "hold_out",
                                           iterations: int = 30,
                                           sample_size: int = -1,
                                           confidence_level: float = 0.05,
                                           seed: int = 42) -> Dict[str, Dict[str, float]]:
        """
        Calculate predictions from embeddings.

        :param embeddings: Iterable or dictionary containing the input embeddings to predict on.
        :param targets: Iterable that contains the targets to calculate metrics
        :param split_name: Name of the split to use for prediction. Default is "hold_out".
        :param iterations: Number of iterations to perform bootstrapping
        :param sample_size: Sample size to use for bootstrapping. -1 defaults to all embeddings which is recommended.
                            It is possible, but not recommended to use a sample size larger or smaller
                            than the number of embeddings, because this might render the variance estimate unreliable.
                            See: https://math.mit.edu/~dav/05.dir/class24-prep-a.pdf (6.2)
        :param confidence_level: Confidence level for result error intervals (0.05 => 95% percentile)
        :param seed: Seed to use for the bootstrapping algorithm
        :return: Dictionary containing the following sub-dictionaries:
                 - 'metrics': Calculated metrics if 'targets' are given, otherwise 'None'.
                 - 'mapped_predictions': Class or value prediction from the given embeddings.
                 - 'mapped_probabilities': Probabilities for classification tasks if include_probabilities is True.
                 Predictions and probabilities are either 'mapped' to keys from an embeddings dict or indexes if
                 embeddings are given as a list.
        """
        if not 0 < confidence_level < 1:
            raise Exception(f"Confidence level must be between 0 and 1, given: {confidence_level}!")

        seed_all(seed)

        if isinstance(embeddings, Dict):
            embeddings_dict = embeddings
        else:
            embeddings_dict = {str(idx): embedding for idx, embedding in enumerate(embeddings)}

        seq_ids = list(embeddings_dict.keys())

        all_predictions = self.from_embeddings(embeddings_dict, targets)["mapped_predictions"]
        all_predictions_dict = self._convert_target_dict(all_predictions)

        all_targets_dict = {seq_id: targets[idx] for idx, seq_id in enumerate(seq_ids)}
        all_targets_dict = self._convert_target_dict(all_targets_dict)

        solver, _ = self.solvers_and_loaders_by_split[split_name]

        return self._do_bootstrapping(iterations=iterations, sample_size=sample_size, confidence_level=confidence_level,
                                      seq_ids=seq_ids, all_predictions_dict=all_predictions_dict,
                                      all_targets_dict=all_targets_dict, metrics_calculator=solver.metrics_calculator)

    @staticmethod
    def _do_bootstrapping(iterations: int,
                          sample_size: int,
                          confidence_level: float,
                          seq_ids: List[str],
                          all_predictions_dict: Dict,
                          all_targets_dict: Dict,
                          metrics_calculator: MetricsCalculator):
        """

        :param iterations: Number of iterations to perform bootstrapping
        :param sample_size: Sample size to use for bootstrapping. -1 defaults to all embeddings which is recommended.
                            It is possible, but not recommended to use a sample size larger or smaller
                            than the number of embeddings, because this might render the variance estimate unreliable.
                            See: https://math.mit.edu/~dav/05.dir/class24-prep-a.pdf (6.2)
        :param confidence_level: Confidence level for result error intervals (0.05 => 95% percentile)
        :param seq_ids: List of sequence IDs
        :param all_predictions_dict: Dictionary of all predictions
        :param all_targets_dict: Dictionary of all targets
        :param metrics_calculator: Metrics calculator object
        :return:
        """
        if sample_size == -1:
            sample_size = len(seq_ids)

        # Convert dictionaries to tensors
        all_predictions = torch.stack([all_predictions_dict[seq_id] for seq_id in seq_ids])
        all_targets = torch.stack([all_targets_dict[seq_id] for seq_id in seq_ids])

        # Set random seed
        seed = np.random.get_state()[1][0] if np.random.get_state() else 42
        rng = np.random.RandomState(seed)

        # Generate all random indices at once
        all_indices = rng.choice(len(seq_ids), size=(iterations, sample_size), replace=True)

        iteration_results = []
        for indices in all_indices:
            # Use integer indexing instead of string keys
            sampled_predictions = all_predictions[indices]
            sampled_targets = all_targets[indices]

            iteration_result = metrics_calculator.compute_metrics(
                predicted=sampled_predictions,
                labels=sampled_targets
            )
            iteration_results.append(iteration_result)

        # Process results
        metrics = list(iteration_results[0].keys())
        result_dict = {}
        for metric in metrics:
            all_metric_values = torch.tensor([res[metric] for res in iteration_results], dtype=torch.float16)
            mean, lower_bound, upper_bound = get_mean_and_confidence_bounds(
                values=all_metric_values,
                dimension=0,
                confidence_level=confidence_level
            )
            result_dict[metric] = {"mean": mean.item(), "lower": lower_bound.item(), "upper": upper_bound.item()}

        return result_dict

    def from_embeddings_with_monte_carlo_dropout(self, embeddings: Union[Iterable, Dict],
                                                 split_name: str = "hold_out",
                                                 n_forward_passes: int = 30,
                                                 confidence_level: float = 0.05,
                                                 seed: int = 42) -> Dict:
        """
        Calculate predictions by using Monte Carlo dropout.
        Only works if the model has at least one dropout layer employed.
        Method to quantify the uncertainty within the model.

        :param embeddings: Iterable or dictionary containing the input embeddings to predict on.
        :param split_name: Name of the split to use for prediction. Default is "hold_out".
        :param n_forward_passes: Number of times to repeat the prediction calculation
                                with different dropout nodes enabled. Must be > 1.
        :param confidence_level: Confidence level for the result confidence intervals. Default is 0.05,
                                which corresponds to a 95% percentile.
        :param seed: Seed to use for the dropout predictions

        :return: Dictionary containing with keys that will either be taken from the embeddings dict or
         represent the indexes if embeddings are given as a list. Contains the following values for each key:
                 - 'prediction': Class or value prediction based on the mean over `n_forward_passes` forward passes.
                 - 'all_predictions': All raw predictions, i.e. one prediction per forward pass.
                 - 'mcd_mean': Average over `n_forward_passes` forward passes for each class.
                 - 'mcd_lower_bound': Lower bound of the confidence interval using a normal distribution with the given
                                      confidence level.
                 - 'mcd_upper_bound': Upper bound of the confidence interval using a normal distribution with the given
                                      confidence level.
                 - 'confidence_range': Confidence range (== mcd_mean - mcd_lower_bound)
        """
        if n_forward_passes <= 1:
            raise ValueError(f"n_forward_passes must be > 1, given: {n_forward_passes}")

        if not 0 < confidence_level < 1:
            raise ValueError(f"Confidence level must be between 0 and 1, given: {confidence_level}!")

        # Necessary because dropout layer have a random part by design
        seed_all(seed)

        solver, dataloader = self._load_solver_and_dataloader(embeddings, split_name)

        predictions = solver.inference_monte_carlo_dropout(dataloader=dataloader,
                                                           n_forward_passes=n_forward_passes,
                                                           confidence_level=confidence_level)["mapped_predictions"]

        # For class predictions, revert from int (model output) to str (class name)
        if self.protocol in Protocol.per_residue_protocols():
            for seq_id, prediction_list in predictions.items():
                for prediction_dict in prediction_list:
                    prediction_dict["prediction"] = list(revert_mappings(protocol=self.protocol,
                                                                         test_predictions={
                                                                             seq_id: prediction_dict["prediction"]
                                                                         },
                                                                         class_int2str=self.class_int2str).values())[0]
        else:
            for seq_id, prediction_dict in predictions.items():
                prediction_dict["prediction"] = list(revert_mappings(protocol=self.protocol,
                                                                     test_predictions={
                                                                         seq_id: prediction_dict["prediction"]
                                                                     },
                                                                     class_int2str=self.class_int2str).values())[0]

        return predictions

    @staticmethod
    def from_onnx_with_embeddings(model_path: str, embeddings: Union[Iterable, Dict],
                                  protocol: Optional[Protocol] = None):
        try:
            import onnxruntime
        except ImportError:
            raise Exception("No onnxruntime in current environment found! Please install one first, e.g. "
                            "uv pip install -e '.[onnx-cpu]'!")

        if isinstance(embeddings, Dict):
            embeddings_dict = embeddings
        else:
            embeddings_dict = {str(idx): embedding for idx, embedding in enumerate(embeddings)}

        onnx_model = onnx.load(model_path)
        onnx.checker.check_model(onnx_model)

        ep_list = ['CUDAExecutionProvider', 'CPUExecutionProvider']
        ort_session = onnxruntime.InferenceSession(model_path, providers=ep_list)

        result_dict = {}
        for seq_id, embedding in embeddings_dict.items():
            input_feed = {ort_session.get_inputs()[0].name: np.expand_dims(embedding, axis=0)}
            ort_outs = ort_session.run(None, input_feed=input_feed)
            if protocol is not None and protocol in Protocol.classification_protocols():
                result_dict[seq_id] = torch.softmax(torch.tensor(ort_outs[0][0]), dim=0).tolist()
            else:
                result_dict[seq_id] = ort_outs[0][0].tolist()[0]
        return result_dict



from .sequence_solvers import SequenceClassificationSolver, SequenceRegressionSolver


# Identical to the SequenceClassificationSolver at the moment
class ResiduesClassificationSolver(SequenceClassificationSolver):
    pass

# Identical to the SequenceRegressionSolver at the moment
class ResiduesRegressionSolver(SequenceRegressionSolver):
    pass


from typing import List, Tuple

from .config_option import ConfigOption, ConfigConstraints, ConfigKey

from ..protocols import Protocol
from ..embedders import get_predefined_embedder_names


def embedding_config(protocol: Protocol) -> Tuple[ConfigKey, List[ConfigOption]]:
    embedding_category = "embedding"
    return ConfigKey.ROOT, [
        ConfigOption(
            name="embedder_name",
            description=f"Allows to define which embedder to use. Can be either a predefined embedder "
                        f"{get_predefined_embedder_names()}, a .onnx model or a huggingface transformer model name.",
            category=embedding_category,
            required=False,
            default="custom_embeddings",
            constraints=ConfigConstraints(
                type=str,
            ),
        ),
        ConfigOption(
            name="custom_tokenizer_config",
            category=embedding_category,
            description="Provide a json config file containing information for a custom tokenizer",
            required=False,
            is_file_option=True,
            default=None,
            constraints=ConfigConstraints(
                type=str,
                allowed_formats=[".json"]
            )
        ),
        ConfigOption(
            name="use_half_precision",
            description="Enable or disable half-precision embedding computations, which can lead to "
                        "performance improvements and reduced memory usage at the potential "
                        "cost of numerical precision. Only applicable for embedding on the GPU.",
            category=embedding_category,
            required=False,
            default=False,
            constraints=ConfigConstraints(
                type=bool,
            )
        ),
        ConfigOption(
            name="embeddings_file",
            category=embedding_category,
            description="Provide a file containing precomputed embeddings, typically in HDF5 (`.h5`) format.",
            required=False,
            is_file_option=True,
            default=None,
            constraints=ConfigConstraints(
                type=str,
                allowed_formats=[".h5", ".hdf5"]
            )
        ),
        ConfigOption(
            name="dimension_reduction_method",
            description="Choose an optional method to reduce embedding dimensions.",
            category=embedding_category,
            required=False,
            default=None,
            constraints=ConfigConstraints(
                type=str,
                allowed_protocols=Protocol.using_per_sequence_embeddings(),
                allowed_values=["umap", "tsne"] if protocol in Protocol.using_per_sequence_embeddings() else []
            )
        ),
        ConfigOption(
            name="n_reduced_components",
            description="Choose the number of reduced components to use for the dimension reduction method.",
            category=embedding_category,
            required=False,
            default=None,
            constraints=ConfigConstraints(
                type=int,
                allowed_protocols=Protocol.using_per_sequence_embeddings(),
                gt=0
            )
        )
    ]



import os
import tempfile

from ruamel import yaml
from pathlib import Path
from biotrainer.config import ConfigurationException
from biotrainer.utilities.cli import train


def test_cross_validation(cv_config: dict):
    print(f"TESTING CONFIG: {cv_config}")
    with tempfile.TemporaryDirectory() as tmp_dir_name:
        tmp_config_path = tmp_dir_name + "/tmp_config_file.yml"
        # Add cv_config to base_config
        with open("test_config.yml", "r") as base_config_file:
            base_config = yaml.safe_load(base_config_file)

        if cv_config["method"] == "k_fold" and cv_config["nested"] is True:
            # Add optimizable hyperparameter for nested k_fold cross validation
            base_config["learning_rate"] = [1e-3, 1e-4]
        else:
            base_config["learning_rate"] = 1e-3
        base_config["cross_validation_config"] = cv_config
        base_config["input_file"] = str(Path("test_input_files/cv_s2v/meltome_cv_test.fasta").absolute())
        base_config["output_dir"] = tmp_dir_name
        base_config["device"] = "cpu"
        with open(tmp_config_path, "w") as tmp_config_file:
            tmp_config_file.write(yaml.dump(base_config))

        try:
            result = train(config=str(Path(tmp_config_path).absolute()))
            assert os.path.exists(f"{tmp_dir_name}/out.yml"), "No output file generated, run failed!"
        except ConfigurationException:
            assert False, "A ConfigurationException was thrown although it shouldn't have."
        except Exception:
            assert False, "An exception was thrown although it shouldn't have."



import torch
import torch.nn as nn


class MaskedMSELoss(nn.Module):
    """ Masked MSE Loss capable of ignoring individual residue/target combinations for residue_to_value protocol """
    def __init__(self, ignore_index: int = -100, reduction: str = 'mean'):
        super().__init__()
        self.ignore_index = ignore_index
        self.reduction = reduction

    def forward(self, input: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        # Create mask for valid (non-ignored) positions
        mask = target != self.ignore_index

        if not mask.any():
            # If all targets are masked, return zero loss
            return torch.tensor(0.0, device=input.device, requires_grad=True)

        # Apply mask to both input and target
        masked_input = input[mask]
        masked_target = target[mask]

        # Compute MSE loss on valid positions only
        loss = nn.functional.mse_loss(masked_input, masked_target, reduction=self.reduction)
        return loss



from biotrainer.embedders import OneHotEncodingEmbedder
from biotrainer.inference import Inferencer

from biotrainer.protocols import Protocol


def onnx_save():
    out_file_path_res = '../residue_to_class/output/out.yml'

    inferencer, out_file = Inferencer.create_from_out_file(out_file_path=out_file_path_res)
    converted_checkpoint_paths = inferencer.convert_to_onnx()
    return converted_checkpoint_paths


def onnx_load(model_path: str):
    # Embed
    sequences = [
        "PROVTEIN",
        "SEQVENCESEQVENCE"
    ]

    embedder = OneHotEncodingEmbedder()
    embeddings = list(embedder.embed_many(sequences))

    # Predict ONNX
    onnx_result = Inferencer.from_onnx_with_embeddings(model_path=model_path, embeddings=embeddings,
                                                       protocol=Protocol.sequence_to_class)
    print(onnx_result)

    # Double check against inferencer predictions
    out_file_path_res = '../residue_to_class/output/out.yml'
    inferencer, out_file = Inferencer.create_from_out_file(out_file_path=out_file_path_res)
    print(inferencer.from_embeddings(embeddings=embeddings, include_probabilities=True))


def main(args=None):
    converted_checkpoint_paths = onnx_save()
    onnx_load(converted_checkpoint_paths[0])


if __name__ == '__main__':
    main()



from pathlib import Path
from typing import Dict, Any, Union, List, Optional

from ...output_files import OutputManager
from ...embedders import PeftEmbeddingService
from ...input_files import BiotrainerSequenceRecord


class PipelineContext:
    """Context object that maintains state throughout the pipeline execution"""

    def __init__(self, config: Dict[str, Any], output_manager: OutputManager, custom_pipeline: bool):
        # Values set prior to pipeline execution
        self.config = config
        self.output_manager = output_manager
        self.custom_pipeline = custom_pipeline

        # Data produced during pipeline execution
        # Setup
        self.pipeline_start_time = None
        self.model_hash = None
        self.hp_manager = None
        # Input Data
        self.input_data: Optional[Union[Path, List[BiotrainerSequenceRecord]]] = None
        # Embedding + Projection
        self.id2emb = None
        self.embedding_service: Optional[PeftEmbeddingService] = None  # For fine-tuning only
        # Data Loading
        self.target_manager = None
        self.n_features = None
        self.n_classes = None
        self.class_str2int: Optional[Dict[str, int]] = None  # Used to apply random_masking
        self.train_dataset = None
        self.val_dataset = None
        self.test_datasets = None
        self.baseline_test_datasets = None  # For random model baseline, uses non-finetuned embeddings
        self.prediction_dataset = None
        self.class_weights = None

        # Training
        self.best_split = None

        # Timing information
        self.pipeline_end_time = None
        self.step_timings = {}



from tqdm import tqdm
from typing import List
from pathlib import Path

from .flip_datasets import FLIP_DATASETS

from ..data_handler import AutoEvalDataHandler, AutoEvalTask

from ...input_files import filter_FASTA


class FLIPDataHandler(AutoEvalDataHandler):
    """Handles FLIP dataset related operations"""

    IGNORE_SPLITS = ["mixed_vs_human_2"]

    @staticmethod
    def get_framework_name() -> str:
        return "FLIP"

    @staticmethod
    def get_download_urls():
        return ["https://nextcloud.cit.tum.de/index.php/s/fqFEeCSpwTHkt8X/download"]

    def preprocess(self, base_path: Path, min_seq_length: int, max_seq_length: int) -> None:
        """ Filters all dataset splits for sequences that fulfill the length requirements """
        for dataset, dataset_info in tqdm(FLIP_DATASETS.items(), desc="Preprocessing datasets"):
            dataset_dir = base_path / dataset

            # Process all splits
            for split in dataset_info["splits"]:
                if split in FLIPDataHandler.IGNORE_SPLITS:
                    continue

                FLIPDataHandler._ensure_preprocessed_file(dataset_dir=dataset_dir,
                                                          name=split,
                                                          min_seq_length=min_seq_length,
                                                          max_seq_length=max_seq_length)

        print("FLIP data preprocessing completed!")

    def get_tasks(self, base_path: Path, min_seq_length: int, max_seq_length: int) -> List[AutoEvalTask]:
        """Build path dictionary for all FLIP datasets"""
        tasks = []

        for dataset, dataset_info in FLIP_DATASETS.items():
            dataset_dir = base_path / dataset
            for split_name in dataset_info["splits"]:
                if split_name in FLIPDataHandler.IGNORE_SPLITS:
                    continue

                input_file = FLIPDataHandler._get_input_file_path(dataset_dir=dataset_dir,
                                                                  name=split_name,
                                                                  min_seq_length=min_seq_length,
                                                                  max_seq_length=max_seq_length)
                if not input_file.exists():
                    raise FileNotFoundError(f"Missing sequence file for {dataset}/{split_name}!")

                tasks.append(
                    AutoEvalTask(name=f"{self.get_framework_name()}-{dataset}-{split_name}", input_file=input_file,
                                 type="Protein"))

        return tasks

    @staticmethod
    def _get_input_file_path(dataset_dir: Path, name: str, min_seq_length: int, max_seq_length: int) -> Path:
        """Get the appropriate input file path (preprocessed if available)"""
        raw_path = dataset_dir / f"{name}.fasta"
        preprocessed_dir_name = FLIPDataHandler.get_preprocessed_file_dir_name(min_seq_length=min_seq_length,
                                                                               max_seq_length=max_seq_length)
        preprocessed_path = dataset_dir / preprocessed_dir_name / f"{name}.fasta"

        if preprocessed_path.exists():
            return preprocessed_path
        return raw_path

    @staticmethod
    def _ensure_preprocessed_file(dataset_dir: Path, name: str, min_seq_length: int, max_seq_length: int) -> Path:
        """Ensure a preprocessed version of the file exists and return its path"""
        download_path = dataset_dir / f"{name}.fasta"
        preprocessed_dir_name = FLIPDataHandler.get_preprocessed_file_dir_name(min_seq_length=min_seq_length,
                                                                               max_seq_length=max_seq_length)
        preprocessed_path = dataset_dir / preprocessed_dir_name / f"{name}.fasta"

        # If preprocessed file already exists, return its path
        if preprocessed_path.exists():
            return preprocessed_path

        # If raw file doesn't exist, we can't proceed
        if not download_path.exists():
            raise FileNotFoundError(f"Required file {download_path} not available for: {name}")

        # Preprocess the file
        preprocessed_dir = dataset_dir / preprocessed_dir_name
        preprocessed_dir.mkdir(exist_ok=True)

        n_kept, n_all = filter_FASTA(input_path=download_path,
                                     output_path=preprocessed_path,
                                     filter_function=lambda seq_record:
                                     min_seq_length <= len(seq_record.seq) <= max_seq_length
                                     )

        print(f"Preprocessed (min: {min_seq_length}, max: {max_seq_length}) {download_path.name}: "
              f"kept {n_kept}/{n_all} sequences")
        return preprocessed_path



import torch
import torch.nn as nn

from .biotrainer_model import BiotrainerModel


# Logistic regression (single linear layer directly mapping to classes)
class LogReg(BiotrainerModel):
    def __init__(self, n_classes: int, n_features: int, **kwargs):
        super(LogReg, self).__init__()
        self.classifier = nn.Sequential(
            nn.Linear(n_features, n_classes),  # 7x32
        )

    def forward(self, x: torch.Tensor, *args, **kwargs) -> torch.Tensor:
        """
            L = protein length
            B = batch-size
            F = number of features (1024 for embeddings)
            N = number of classes (9 for conservation)
        """
        Yhat = self.classifier(x)  # IN: X = (B x L x F); OUT: Yhat_consurf = (B x L x N)
        return Yhat



import numpy as np

from typing import Any, Dict, List, Optional
from sklearn.model_selection import KFold, StratifiedKFold, RepeatedStratifiedKFold, RepeatedKFold, LeavePOut

from ..protocols import Protocol
from ..utilities import Split, get_logger

logger = get_logger(__name__)


class CrossValidationSplitter:

    def __init__(self, protocol: Protocol, cross_validation_config: Dict[str, Any]):
        self._protocol = protocol

        self._split_strategy = None
        if cross_validation_config["method"] == "hold_out":
            self._split_strategy = self._hold_out_split

        if cross_validation_config["method"] == "k_fold":
            k = int(cross_validation_config["k"])
            if "stratified" in cross_validation_config.keys():
                stratified = eval(str(cross_validation_config["stratified"]).capitalize())
            else:
                stratified = False
            if "repeat" in cross_validation_config.keys():
                repeat = int(cross_validation_config["repeat"])
            else:
                repeat = 1
            if "nested_k" in cross_validation_config.keys():
                nested_k = int(cross_validation_config["nested_k"])
                self._nested_split_strategy = lambda train_dataset, current_outer_k, hp_iteration: \
                    self._k_fold_split(k=nested_k, stratified=stratified, nested=True, repeat=repeat,
                                       train_dataset=train_dataset, val_dataset=[],
                                       current_outer_k=current_outer_k, hp_iteration=hp_iteration)
            self._split_strategy = lambda train_dataset, val_dataset: \
                self._k_fold_split(k=k, stratified=stratified, nested=False, repeat=repeat, train_dataset=train_dataset,
                                   val_dataset=val_dataset)

        if cross_validation_config["method"] == "leave_p_out":
            p = int(cross_validation_config["p"])
            self._split_strategy = lambda train_dataset, val_dataset: \
                self._leave_p_out_split(p=p, train_dataset=train_dataset, val_dataset=val_dataset)

    def split(self, train_dataset: List, val_dataset: List) -> List[Split]:
        return self._split_strategy(train_dataset, val_dataset)

    def nested_split(self, train_dataset: List, current_outer_k: int, hp_iteration: int) -> List[Split]:
        return self._nested_split_strategy(train_dataset, current_outer_k, hp_iteration)

    @staticmethod
    def _hold_out_split(train_dataset: List, val_dataset: List) -> List[Split]:
        return [Split("hold_out", train_dataset, val_dataset)]

    @staticmethod
    def _continuous_values_to_bins(ys: List, number_bins: Optional[int] = 10) -> List:
        """
        Calculate bins for continuous target values.
        Assigns to each data_point in ys the maximum of its associated bin.
        Enables stratified k-fold cross validation for x_to_value protocols.
            params:
                ys: continuous targets
                number_bins: number of bins to create
            returns: List of bins (len == len of ys)
        """
        ys_np = np.array(ys)
        maximum = np.max(ys_np)
        minimum = np.min(ys_np)

        bin_size = abs(maximum - minimum) / float(number_bins)

        bins = []
        bin_dict = {}  # For logging
        bin_min = minimum
        bin_max = bin_min + bin_size
        for data_point in np.sort(ys_np):
            if bin_min <= data_point <= bin_max:
                if str(bin_max) not in bin_dict.keys():
                    bin_dict[str(bin_max)] = 0

                bins.append(str(bin_max))
                bin_dict[str(bin_max)] += 1
            else:
                while data_point > bin_max:
                    bin_min += bin_size
                    bin_max += bin_size
                    if str(bin_max) not in bin_dict.keys():
                        bin_dict[str(bin_max)] = 0

                bins.append(str(bin_max))
                bin_dict[str(bin_max)] += 1

        if len(bins) != len(ys) or sum(bin_dict.values()) != len(ys):
            raise Exception(f"Could not correctly calculate bins for continuous values! "
                            f"(Number bins {len(bins)} != Number targets {len(ys)})\n"
                            f"Consider setting stratified to False or filing an issue.")

        logger.info("Transformed continuous targets to bins for Cross Validation split, statistics:")
        for idx, bin_value in enumerate(sorted(set(bins))):
            logger.info(f"{idx}: {bin_value}: {bin_dict[bin_value]} ({100 * bin_dict[bin_value] / len(bins)} %)")

        return bins

    def _k_fold_split(self, k: int, stratified: bool, nested: bool, repeat: int,
                      train_dataset: List, val_dataset: List,
                      hp_iteration: Optional[int] = None,
                      current_outer_k: Optional[int] = None) -> List[Split]:
        concat_dataset = train_dataset + val_dataset
        ys = [sample.target for sample in concat_dataset]

        split_base_name = "k_fold"
        inner_or_outer_splits = "(inner split)" if not nested else "(outer splits)"
        if stratified:
            logger.info(f"Splitting to stratified {k}-fold Cross Validation datasets {inner_or_outer_splits}")
            split_base_name += "-strat"
            if repeat > 1:
                split_base_name += "-rep"
                kf = RepeatedStratifiedKFold(n_splits=k, n_repeats=repeat)
            else:
                kf = StratifiedKFold(n_splits=k)
            # Change continuous values to bins for stratified split
            if self._protocol in Protocol.regression_protocols():
                ys = self._continuous_values_to_bins(ys)
        else:
            logger.info(f"Splitting to {k}-fold Cross Validation datasets {inner_or_outer_splits}")
            if repeat > 1:
                split_base_name += "-rep"
                kf = RepeatedKFold(n_splits=k, n_repeats=repeat)
            else:
                kf = KFold(n_splits=k)
        if current_outer_k:
            split_base_name += f"-{current_outer_k}"
        if hp_iteration:
            split_base_name += f"-hp-{hp_iteration}"
        if nested:
            split_base_name += "-nested"
        all_splits = []
        for idx, (split_ids_train, split_ids_val) in enumerate(kf.split(X=concat_dataset, y=ys)):
            train_split = [concat_dataset[split_id_train] for split_id_train in split_ids_train]
            val_split = [concat_dataset[split_id_val] for split_id_val in split_ids_val]

            if repeat > 1:
                # Repeated splits are labeled e.g. 1-1, 1-2, 2-1, 2-2, 3-1, 3-2 .. (for 3==2 and repeat==2)
                current_repeated_split = f"{(idx + 1) % k if (idx + 1) % k != 0 else k}-{(idx // k) + 1}"
                current_split_name = f"{split_base_name}-{current_repeated_split}"
            else:
                current_split_name = f"{split_base_name}-{idx + 1}"

            all_splits.append(Split(current_split_name, train_split, val_split))

        return all_splits

    @staticmethod
    def _leave_p_out_split(p: int,
                           train_dataset: List, val_dataset: List) -> List[Split]:
        concat_dataset = train_dataset + val_dataset

        lpo = LeavePOut(p=p)
        n_created_datasets = lpo.get_n_splits(X=concat_dataset)
        logger.info(f"Splitting to leave_{p}_out Cross Validation datasets")
        logger.info(f"Number of created datasets: {n_created_datasets}")
        if n_created_datasets > 1000:
            logger.warning(f"Number of created datasets is very high! Consider using a smaller value for p or another "
                           f"cross validation method.")

        all_splits = []
        for idx, (split_ids_train, split_ids_val) in enumerate(lpo.split(X=concat_dataset)):
            train_split = [concat_dataset[split_id_train] for split_id_train in split_ids_train]
            val_split = [concat_dataset[split_id_val] for split_id_val in split_ids_val]
            all_splits.append(Split(f"leave_{p}_out-{idx + 1}", train_split, val_split))

        return all_splits



from typing import Dict, List

from .pipeline_context import PipelineContext
from .pipeline_step import PipelineStep, PipelineStepType

from ...utilities import get_logger

logger = get_logger(__name__)


class Pipeline:
    def __init__(self):
        self.steps_dict: Dict[PipelineStepType, PipelineStep] = {}
        self.step_order: List[PipelineStepType] = []

    def add_step(self, step: PipelineStep):
        step_type = step.get_step_type()

        self.steps_dict[step_type] = step
        if step_type not in self.step_order:
            self.step_order.append(step_type)
        return self

    def replace_step(self, step: PipelineStep):
        step_type = step.get_step_type()

        if step_type not in self.steps_dict:
            raise ValueError(f"Step type {step_type} not found in pipeline")
        self.steps_dict[step_type] = step
        return self

    def execute(self, context: PipelineContext) -> PipelineContext:
        current_context = context
        for step_type in self.step_order:
            step = self.steps_dict[step_type]
            try:
                logger.info(f"Executing step: {step.__class__.__name__}")
                current_context = step.process(current_context)
            except Exception as e:
                logger.error(f"Error in step {step.__class__.__name__}: {str(e)}")
                raise
        return current_context



import torch
import itertools

from pathlib import Path
from collections import Counter
from typing import Dict, Any, Tuple, Optional, List, Union

from ..protocols import Protocol
from ..input_files import BiotrainerSequenceRecord, merge_protein_interactions, get_split_lists, read_FASTA
from ..utilities import MASK_AND_LABELS_PAD_VALUE, INTERACTION_INDICATOR, EmbeddingDatasetSample, get_logger, \
    SequenceDatasetSample

logger = get_logger(__name__)


class TargetManager:
    _input_records: Dict[str, BiotrainerSequenceRecord]  # Hash To Record (for Fine-Tuning)
    _input_ids: Dict[str, List[str]] = dict()  # Hash To List of IDs
    _id2target: Dict[str, Any] = dict()  # Hash To Target
    _id2sets: Dict[str, str] = dict()  # Hash to Set
    # This will be 1 for regression tasks, 2 for binary classification tasks, and N>2 for everything else
    number_of_outputs: int = 1

    # Optional for classification tasks, must be set in _calculate_targets()
    class_str2int: Optional[Dict[str, int]] = None
    class_int2str: Optional[Dict[int, str]] = None
    _class_labels: Optional[List[str]] = None

    # Dataset split lists
    training_ids: List[str] = None
    validation_ids: List[str] = None
    testing_ids: Dict[str, List[str]] = None
    prediction_ids: List[str] = None

    # Interaction operations
    interaction_operations = {
        "multiply": lambda embedding_left, embedding_right: torch.mul(embedding_left, embedding_right),
        "concat": lambda embedding_left, embedding_right: torch.concat([embedding_left, embedding_right])
    }

    def __init__(self, protocol: Protocol, input_data: Union[str, Path, List[BiotrainerSequenceRecord]],
                 ignore_file_inconsistencies: Optional[bool] = False,
                 cross_validation_method: str = "",
                 interaction: Optional[str] = None):
        self.protocol = protocol
        self._input_data = input_data
        self._ignore_file_inconsistencies = ignore_file_inconsistencies
        self._cross_validation_method = cross_validation_method
        self._interaction = interaction

    def _calculate_targets(self):
        # Parse FASTA protein sequences if not done yet
        if isinstance(self._input_data, str) or isinstance(self._input_data, Path):
            input_records: List[BiotrainerSequenceRecord] = read_FASTA(self._input_data)
        else:
            input_records: List[BiotrainerSequenceRecord] = self._input_data
        assert isinstance(input_records[0], BiotrainerSequenceRecord)

        # Store records for fine-tuning sequence datasets
        self._input_records = {seq_record.get_hash(): seq_record for seq_record in input_records}

        # Store input ids for better error messages
        for seq_record in input_records:
            seq_hash = seq_record.get_hash()
            if seq_hash not in self._input_ids:
                self._input_ids[seq_hash] = []
            self._input_ids[seq_hash].append(seq_record.seq_id)

        # id2X => sequence hash to X
        self._id2target, id2masks, self._id2sets = BiotrainerSequenceRecord.get_dicts(input_records)
        assert len(self._id2target.keys()) == len(id2masks.keys()) \
               == len(self._id2sets.keys()), f"Length mismatch after reading input file!"

        if self._interaction:
            merged_interactions_dict = merge_protein_interactions(input_records)
            self._id2target = {int_id: int_dict["TARGET"] for int_id, int_dict in
                               merged_interactions_dict.items()}
            self._id2sets = {int_id: int_dict["SET"] for int_id, int_dict in
                             merged_interactions_dict.items()}

        # Remove None targets from pred set
        self._id2target = {seq_hash: target for seq_hash, target in self._id2target.items() if target is not None}
        if len(self._id2target) == 0:
            raise ValueError("Could not parse any valid targets from given input file!")

        # 1. Residue Level
        if self.protocol in Protocol.per_residue_protocols():
            # a) Class output
            if self.protocol in Protocol.classification_protocols():
                class_labels_temp = set()

                # Infer classes from data
                for classes in self._id2target.values():
                    class_labels_temp = class_labels_temp | set(classes)
                self._class_labels = sorted(class_labels_temp)

                self.number_of_outputs = len(self._class_labels)

                # Create a mapping from integers to class labels and reverse
                self.class_str2int = {letter: idx for idx, letter in enumerate(self._class_labels)}
                self.class_int2str = {idx: letter for idx, letter in enumerate(self._class_labels)}

                # Convert label values to lists of numbers based on the maps
                self._id2target = {identifier: torch.tensor([self.class_str2int[label] for label in labels]).long()
                                   for identifier, labels in self._id2target.items()}  # classes idxs (zero-based)

                # MASKS
                self._apply_masks(id2masks=id2masks)
            # b) Value output
            elif self.protocol in Protocol.regression_protocols():
                self._id2target = {seq_hash: torch.tensor(res_val).float()
                                   for seq_hash, res_val in self._id2target.items()}
                self.number_of_outputs = 1
                self._apply_masks(id2masks=id2masks)
            else:
                raise NotImplementedError

        # 2. Sequence Level
        elif self.protocol in Protocol.per_sequence_protocols():
            # a) Class output
            if self.protocol in Protocol.classification_protocols():
                # Infer classes from data
                self._class_labels = sorted(set(self._id2target.values()))
                self.number_of_outputs = len(self._class_labels)

                # Create a mapping from integers to class labels and reverse
                self.class_str2int = {letter: idx for idx, letter in enumerate(self._class_labels)}
                self.class_int2str = {idx: letter for idx, letter in enumerate(self._class_labels)}

                # Convert label values to lists of numbers based on the maps
                self._id2target = {seq_hash: torch.tensor(self.class_str2int[label]).long()
                                   for seq_hash, label in self._id2target.items()}  # classes idxs (zero-based)

            # b) Value output
            elif self.protocol in Protocol.regression_protocols():
                self._id2target = {seq_hash: torch.tensor(float(seq_val)).float()
                                   for seq_hash, seq_val in self._id2target.items()}
                self.number_of_outputs = 1
            else:
                raise NotImplementedError
        else:
            raise NotImplementedError

        # Double-checking after target loading for correctness
        if self.protocol in Protocol.regression_protocols():
            assert self.number_of_outputs == 1, f"Number of outputs does not equal 1 for regression protocol!"
        if not self._id2target or len(self._id2target) == 0:
            raise ValueError("Prediction targets not found or could not be extracted!")

    def _apply_masks(self, id2masks):
        # Validate masks (each mask must contain at least one resolved (1) value)
        for seq_hash, mask in id2masks.items():
            if mask is not None and 1 not in mask:
                raise ValueError(f"{self._input_ids[seq_hash]} does not have a valid mask as it does "
                                 f"not contain at least one resolved (1) value!")

        # Replace labels with masking value
        for seq_hash, unmasked_target in self._id2target.items():
            mask = id2masks[seq_hash]
            if mask is not None:
                target_with_mask = torch.tensor([value if mask[index] == 1 else MASK_AND_LABELS_PAD_VALUE for
                                                 index, value in enumerate(unmasked_target)])
                target_with_mask = target_with_mask.long() if self.protocol in Protocol.classification_protocols() \
                    else target_with_mask.float()
                self._id2target[seq_hash] = target_with_mask

    def _validate_targets(self, id2emb: Dict[str, torch.tensor]):
        """
        1. Check if number of embeddings and corresponding labels match for every protocol:
            # embeddings == # labels --> SUCCESS
            # embeddings > # labels --> Fail, unless self._ignore_file_inconsistencies (=> drop embeddings).
            # embeddings < # labels --> Fail, unless self._ignore_file_inconsistencies (=> drop labels).
        2. Check that lengths of embeddings == length of provided labels, if not --> Fail. (only residue_to_x)
        """
        invalid_embeddings_lengths = []
        embeddings_without_labels = []
        labels_without_embeddings = []
        if self._interaction:
            all_protein_ids = []
            for interaction_key in self._id2target.keys():
                protein_ids = interaction_key.split(INTERACTION_INDICATOR)
                all_protein_ids.extend(protein_ids)
            all_hashes_with_target = set(all_protein_ids)
        else:
            all_hashes_with_target = set(self._id2target.keys())

        for seq_hash, embd in id2emb.items():
            # Check that all embeddings have a corresponding label
            if seq_hash in self.prediction_ids:
                continue

            if seq_hash not in all_hashes_with_target:
                embeddings_without_labels.append(seq_hash)
            # Make sure the length of the sequences in the embeddings match the length of the seqs in the labels
            elif self.protocol in Protocol.per_residue_protocols() and len(embd) != self._id2target[seq_hash].size()[0]:
                invalid_embeddings_lengths.append((seq_hash, len(embd), self._id2target[seq_hash].size()))

        for seq_hash in all_hashes_with_target:
            # Check that all labels have a corresponding embedding
            if seq_hash not in id2emb.keys():
                labels_without_embeddings.append(seq_hash)

        if len(embeddings_without_labels) > 0:
            if self._ignore_file_inconsistencies:
                if len(embeddings_without_labels) == len(id2emb):
                    raise ValueError(f"Did not find any labels for any given embedding!")

                logger.warning(f"Found {len(embeddings_without_labels)} embedding(s) without a corresponding "
                               f"entry in the labels file! Because ignore_file_inconsistencies flag is set, "
                               f"these sequences are dropped for training. "
                               f"Data loss: {(len(embeddings_without_labels) / len(id2emb.keys())) * 100:.2f}%")
                for seq_hash in embeddings_without_labels:
                    id2emb.pop(seq_hash)  # Remove redundant sequences
            else:
                exception_message = f"{len(embeddings_without_labels)} sequence(s) not found in labels file! " \
                                    f"Make sure that all sequences are present and annotated in the labels file " \
                                    f"or set the ignore_file_inconsistencies flag to True.\n" \
                                    f"Missing label sequence ids:\n"
                for seq_hash in embeddings_without_labels:
                    exception_message += f"Sequence - ID(s): {self._input_ids.get(seq_hash, 'N/A')} - Hash: {seq_hash}\n"
                raise ValueError(exception_message[:-1])  # Discard last \n

        if len(labels_without_embeddings) > 0:

            if self._ignore_file_inconsistencies:
                logger.warning(f"Found {len(labels_without_embeddings)} label(s) without a corresponding "
                               f"entry in the embeddings file! Because ignore_file_inconsistencies flag is set, "
                               f"these labels are dropped for training. "
                               f"Data loss: {(len(labels_without_embeddings) / len(id2emb.keys())) * 100:.2f}%")
                for seq_hash in labels_without_embeddings:
                    self._id2target.pop(seq_hash)  # Remove redundant labels
            else:
                exception_message = f"{len(labels_without_embeddings)} label(s) not found in embeddings file! " \
                                    f"Make sure that for every sequence id in the labels file, there is a " \
                                    f"corresponding embedding. If no pre-computed embeddings were used, \n" \
                                    f"this error message indicates that there is a sequence missing in the " \
                                    f"sequence_file.\n" \
                                    f"Setting the ignore_file_inconsistencies flag to True " \
                                    f"will ignore this problem.\n" \
                                    f"Missing sequence ids:\n"
                for seq_hash in labels_without_embeddings:
                    exception_message += f"Sequence - ID(s): {self._input_ids.get(seq_hash, 'N/A')} - Hash: {seq_hash}\n"
                raise ValueError(exception_message[:-1])  # Discard last \n

        if len(invalid_embeddings_lengths) > 0:
            exception_message = f"Length mismatch for {len(invalid_embeddings_lengths)} embedding(s)!\n"
            for seq_hash, seq_len, target_len in invalid_embeddings_lengths:
                exception_message += (
                    f"Embedding - ID(s): {self._input_ids.get(seq_hash, 'N/A')} - Sequence Hash: {seq_hash} - "
                    f"Embedding_Length={seq_len} vs. Labels_Length={self._id2target[seq_hash].size()[0]}\n")
            raise ValueError(exception_message[:-1])  # Discard last \n

    @staticmethod
    def _validate_embeddings_shapes(id2emb: Dict[str, Any]):
        shapes = set([val.shape[-1] for val in id2emb.values()])
        all_embeddings_have_same_dimension = len(shapes) == 1
        if not all_embeddings_have_same_dimension:
            raise ValueError(f"Embeddings dimensions differ between sequences, but all must be equal!\n"
                             f"Found: {shapes}")

    def get_embedding_datasets(self, id2emb: Dict[str, Any]) -> \
            Tuple[List[EmbeddingDatasetSample], List[EmbeddingDatasetSample], Dict[str, List[EmbeddingDatasetSample]],
            List[EmbeddingDatasetSample]]:
        self._calculate_targets()

        # Get dataset splits from file
        self.training_ids, self.validation_ids, self.testing_ids, self.prediction_ids = get_split_lists(self._id2sets)

        self._validate_targets(id2emb)

        # Check dataset splits are not empty
        def except_on_empty(split_ids: List[str], name: str):
            if len(split_ids) == 0:
                raise ValueError(f"The provided {name} set is empty! Please provide at least one sequence for "
                                 f"the {name} set.")

        if not self._ignore_file_inconsistencies:
            except_on_empty(split_ids=self.training_ids, name="training")
            if self._cross_validation_method == "hold_out":
                except_on_empty(split_ids=self.validation_ids, name="validation")
            for test_set in self.testing_ids.values():
                except_on_empty(split_ids=test_set, name="test")

        # Combine embeddings for protein_protein_interaction
        if self._interaction:
            interaction_operation = self.interaction_operations.get(self._interaction)
            if not interaction_operation:
                raise NotImplementedError(f"Chosen interaction operation {self._interaction} is not supported!")
            interaction_id2emb = {}
            for interaction_id in self._id2target.keys():
                interactor_left = interaction_id.split(INTERACTION_INDICATOR)[0]
                interactor_right = interaction_id.split(INTERACTION_INDICATOR)[1]
                embedding_left = id2emb[interactor_left]
                embedding_right = id2emb[interactor_right]
                combined_embedding = interaction_operation(embedding_left, embedding_right)
                interaction_id2emb[interaction_id] = combined_embedding

            id2emb = interaction_id2emb

        # Validate that all embeddings have the same shape
        self._validate_embeddings_shapes(id2emb)

        # Create datasets
        train_dataset = [
            EmbeddingDatasetSample(idx, id2emb[idx], torch.tensor(self._id2target[idx])) for idx in self.training_ids
        ]
        val_dataset = [
            EmbeddingDatasetSample(idx, id2emb[idx], torch.tensor(self._id2target[idx])) for idx in self.validation_ids
        ]

        test_datasets = {}
        for test_set_id, test_set in self.testing_ids.items():
            test_datasets[test_set_id] = [EmbeddingDatasetSample(idx, id2emb[idx], torch.tensor(self._id2target[idx]))
                                          for idx in test_set]

        pred_dataset = [
            EmbeddingDatasetSample(idx, id2emb[idx], torch.empty(1)) for idx in self.prediction_ids
        ]

        return train_dataset, val_dataset, test_datasets, pred_dataset

    def get_sequence_datasets(self):
        self._calculate_targets()

        # Get dataset splits from file
        self.training_ids, self.validation_ids, self.testing_ids, self.prediction_ids = get_split_lists(self._id2sets)

        # TODO DUPLICATED CODE
        # Check dataset splits are not empty
        def except_on_empty(split_ids: List[str], name: str):
            if len(split_ids) == 0:
                raise ValueError(f"The provided {name} set is empty! Please provide at least one sequence for "
                                 f"the {name} set.")

        if not self._ignore_file_inconsistencies:
            except_on_empty(split_ids=self.training_ids, name="training")
            if self._cross_validation_method == "hold_out":
                except_on_empty(split_ids=self.validation_ids, name="validation")
            for test_set in self.testing_ids.values():
                except_on_empty(split_ids=test_set, name="test")

        # Create datasets
        train_dataset = [SequenceDatasetSample(idx, self._input_records[idx], self._id2target[idx])
                         for idx in self.training_ids]
        val_dataset = [SequenceDatasetSample(idx, self._input_records[idx], self._id2target[idx])
                       for idx in self.validation_ids]

        test_datasets = {}
        for test_set_id, test_set in self.testing_ids.items():
            test_datasets[test_set_id] = [SequenceDatasetSample(idx, self._input_records[idx], self._id2target[idx])
                                          for idx in test_set]

        pred_dataset = [SequenceDatasetSample(idx, self._input_records[idx], self._id2target[idx])
                        for idx in self.prediction_ids]

        return train_dataset, val_dataset, test_datasets, pred_dataset

    def compute_class_weights(self) -> torch.FloatTensor:
        if self.protocol in Protocol.classification_protocols():
            training_targets = [self._id2target[training_id] for training_id in self.training_ids]
            if self.protocol in Protocol.per_residue_protocols():
                # concatenate all targets irrespective of protein to count class sizes
                training_targets = list(itertools.chain.
                                        from_iterable([list(targets) for targets in training_targets]))
                # Ignore unresolved residues for class weights
                training_targets = [target.item() for target in training_targets if target != MASK_AND_LABELS_PAD_VALUE]

                counter = Counter(training_targets)
            else:  # Per-sequence
                counter = Counter([target.item() for target in training_targets])
            # total number of samples in the set irrespective of classes
            n_samples = sum([counter[idx] for idx in range(len(self.class_str2int))])
            n_classes = len(counter.keys())
            # balanced class weighting (inversely proportional to class size)
            class_weights = [(n_samples / (n_classes * counter[idx])) if counter[idx] > 0 else 1
                             for idx in range(len(self.class_str2int))]

            logger.info(f"Total number of sequences/residues: {n_samples}")
            logger.info("Individual class counts and weights (training set):")
            for idx in range(len(self.class_str2int)):
                logger.info(f"\t{self.class_int2str[idx]} : {counter[idx]} ({class_weights[idx]:.3f})")
            if n_classes < len(self.class_str2int):
                missing_classes = [self.class_int2str[idx] for idx in range(len(self.class_str2int))
                                   if counter[idx] == 0]
                logger.warning(f"The training set does not contain samples "
                               f"for the following classes: {missing_classes}!")
            return torch.FloatTensor(class_weights)
        else:
            raise ValueError(f"Class weights can only be calculated for classification tasks!")



import time
import datetime

from pathlib import Path

from ..hp_manager import HyperParameterManager
from ..pipeline import PipelineStep, PipelineContext
from ..pipeline.pipeline_step import PipelineStepType

from ...protocols import Protocol
from ...utilities import seed_all, get_logger, __version__, setup_logging, get_device, calculate_model_hash

logger = get_logger(__name__)


class SetupStep(PipelineStep):

    def get_step_type(self) -> PipelineStepType:
        return PipelineStepType.SETUP

    @staticmethod
    def _post_process_config(context: PipelineContext):
        context.config["protocol"] = Protocol.from_string(context.config["protocol"])

        # Create output dir
        output_dir = Path(context.config["output_dir"])
        if not output_dir.is_dir():
            output_dir.mkdir(parents=True)
        context.config["output_dir"] = output_dir

        # Setup logging
        setup_logging(str(output_dir), context.config["num_epochs"])

        # Create log directory (if necessary)
        embedder_name = context.config["embedder_name"].split("/")[-1]
        log_dir = output_dir / context.config["model_choice"] / embedder_name
        if not log_dir.is_dir():
            logger.info(f"Creating log-directory: {log_dir}")
            log_dir.mkdir(parents=True)
        context.config["log_dir"] = str(log_dir)

        # Get device once at the beginning
        device = get_device(context.config["device"] if "device" in context.config.keys() else None)
        context.config["device"] = device

        # Set input data
        context.input_data = context.config["input_file"]

    def process(self, context: PipelineContext) -> PipelineContext:
        context.pipeline_start_time = time.perf_counter()
        pipeline_start_time_abs = str(datetime.datetime.now().isoformat())

        self._post_process_config(context)

        # Log version
        logger.info(f"** Running biotrainer (v{__version__}) training routine **")
        context.output_manager.add_derived_values({'biotrainer_version': str(__version__)})
        # Log start time
        logger.info(f"Pipeline start time: {pipeline_start_time_abs}")
        context.output_manager.add_derived_values({'pipeline_start_time': pipeline_start_time_abs})

        if "pretrained_model" in context.config.keys():
            logger.info(f"Using pre_trained model: {context.config['pretrained_model']}")

        # Create hyperparameter manager
        hp_manager = HyperParameterManager(**context.config)
        context.hp_manager = hp_manager

        # Calculate model hash
        model_hash = calculate_model_hash(dataset_files=[Path(val) for key, val in context.config.items()
                                                         if "_file" in key and Path(str(val)).exists()],
                                          config=context.config,
                                          custom_trainer=context.custom_pipeline
                                          )
        context.model_hash = model_hash
        logger.info(f"Training model with hash: {model_hash}")
        context.output_manager.add_derived_values({"model_hash": model_hash})
        # Seed
        seed = context.config["seed"]
        seed_all(seed)
        logger.info(f"Using seed: {seed}")
        # Log device
        logger.info(f"Using device: {context.config['device']}")

        context.output_manager.add_config(context.config)
        return context


import torch

from torch.utils.data import DataLoader
from contextlib import nullcontext as _nullcontext
from typing import Dict, Union, Optional, Callable, List

from .solver import Solver
from .solver_utils import get_mean_and_confidence_bounds


class ResidueSolver(Solver):
    def _transform_network_output(self, network_output: torch.Tensor) -> torch.Tensor:
        # TODO [Refactoring] Optimize transform detection, maybe put into model class itself
        downstream_network = self.network.get_downstream_model()

        if isinstance(downstream_network, torch._dynamo.eval_frame.OptimizedModule):
            network_type = type(downstream_network._orig_mod).__name__
        else:
            network_type = type(downstream_network).__name__

        if network_type in ["FNN", "DeeperFNN", "LogReg"]:
            # (Batch_size x protein_Length x Number_Outputs) => (B x N x L)
            network_output = network_output.permute(0, 2, 1)

        return network_output

    # Gets overwritten to shorten prediction lengths if necessary
    def _training_iteration(
            self, x: torch.Tensor, y: torch.Tensor, step=1, context: Optional[Callable] = None,
            lengths: Optional[torch.LongTensor] = None
    ) -> Dict[str, Union[float, list, Dict[str, Union[float, int]]]]:
        result_dict = super()._training_iteration(x, y, step, context, lengths)
        if not context:
            context = _nullcontext

        with context():
            prediction = result_dict['prediction']
            # If lengths is defined, we need to shorten the residue predictions to the length
            if lengths is not None:
                return_pred = list()
                for pred_x, length_x in zip(prediction, lengths):
                    return_pred.append(pred_x[:length_x])

                result_dict['prediction'] = return_pred

            return result_dict

    # Gets overwritten to shorten prediction lengths if necessary
    def _prediction_iteration(self, x: torch.Tensor, lengths: Optional[torch.LongTensor] = None) -> Dict[str, List]:
        result_dict = super()._prediction_iteration(x, lengths)
        with torch.no_grad():
            predictions = result_dict['prediction']
            probabilities = result_dict['probabilities']
            # If lengths is defined, we need to shorten the residue predictions and probabilities to the length
            if lengths is not None:
                shortened_predictions = []
                shortened_probabilities = []
                for original_prediction, original_probability, length_to_shorten in zip(predictions, probabilities,
                                                                                        lengths):
                    shortened_predictions.append(original_prediction[:length_to_shorten])

                    if isinstance(original_probability[0], float):  # Regression
                        shortened_probabilities.append(original_probability[:length_to_shorten])
                    else:  # Classification => List of output probabilities
                        shortened_output_probabilities = []
                        for per_output_probabilities in original_probability:  # len(original_probability) == 1 for regression
                            shortened_output_probabilities.append(per_output_probabilities[:length_to_shorten])
                        shortened_probabilities.append(shortened_output_probabilities)

                result_dict['prediction'] = shortened_predictions
                result_dict['probabilities'] = shortened_probabilities

            return result_dict

    def inference_monte_carlo_dropout(self, dataloader: DataLoader,
                                      n_forward_passes: int = 30,
                                      confidence_level: float = 0.05):
        """
        Calculate inference results from existing models for given embeddings.
        Adaption needed for residue_to_x tasks because of multiple predictions for each sequence

            dataloader: Dataloader with embeddings
            n_forward_passes: Times to repeat calculation with different dropout nodes enabled
            confidence_level: Confidence level for result confidence intervals (0.05 => 95% percentile)
        """
        if not 0 < confidence_level < 1:
            raise ValueError(f"Confidence level must be between 0 and 1, given: {confidence_level}!")

        mapped_predictions = {}
        is_regression = isinstance(self, ResidueRegressionSolver)

        for i, (seq_ids, X, y, lengths) in enumerate(dataloader):
            dropout_iterations = self._do_dropout_iterations(X, lengths, n_forward_passes)

            # Get outputs individually for each sequence
            outputs_by_sequence = {}
            for dropout_iteration in dropout_iterations:
                dropout_outputs = dropout_iteration["probabilities"]  # This contains raw outputs for regression
                for idx, outputs in enumerate(dropout_outputs):
                    if seq_ids[idx] not in outputs_by_sequence.keys():
                        outputs_by_sequence[seq_ids[idx]] = []
                    outputs_by_sequence[seq_ids[idx]].append(outputs)

            # Calculate dropout mean and confidence range for each residue in sequence
            seq_idx = 0
            for seq_id, dropout_residues in outputs_by_sequence.items():
                stacked_residues_tensor = torch.stack([torch.tensor(outputs) for outputs in dropout_residues], dim=1)

                dropout_mean, lower_bound, upper_bound = get_mean_and_confidence_bounds(
                    values=stacked_residues_tensor,
                    dimension=1,
                    confidence_level=confidence_level
                )

                # Different prediction logic for classification vs regression
                if is_regression:
                    # For regression, the mean is the prediction
                    predictions_by_mean = dropout_mean.squeeze()
                else:
                    # For classification, take argmax of mean probabilities
                    _, predictions_by_mean = torch.max(dropout_mean, dim=0)

                # Create dict with seq_id: prediction
                mapped_predictions[seq_id] = []
                for residue_idx, residue_prediction in enumerate(predictions_by_mean):
                    mapped_predictions[seq_id].append({
                        "prediction": residue_prediction.item(),
                        "all_predictions": [dropout_iteration["prediction"][seq_idx][residue_idx] for
                                            dropout_iteration in dropout_iterations],
                        "mcd_mean": dropout_mean.T[residue_idx] if not is_regression else dropout_mean[residue_idx],
                        "mcd_lower_bound": lower_bound.T[residue_idx] if not is_regression else lower_bound[
                            residue_idx],
                        "mcd_upper_bound": upper_bound.T[residue_idx] if not is_regression else upper_bound[
                            residue_idx],
                    })

                seq_idx += 1

        return {'mapped_predictions': mapped_predictions}


class ResidueClassificationSolver(ResidueSolver):

    def _logits_to_probabilities(self, logits: torch.Tensor) -> torch.Tensor:
        return torch.softmax(logits, dim=1)

    def _probabilities_to_predictions(self, probabilities: torch.Tensor) -> torch.Tensor:
        _, predicted_classes = torch.max(probabilities, dim=1)

        return predicted_classes


class ResidueRegressionSolver(ResidueSolver):
    def _transform_network_output(self, network_output: torch.Tensor) -> torch.Tensor:
        network_output = super()._transform_network_output(network_output)
        network_output = network_output.squeeze(1)
        return network_output



import biotrainer.datasets
import biotrainer.losses
import biotrainer.models
import biotrainer.optimizers
import biotrainer.solvers
import biotrainer.trainers
import biotrainer.utilities
import biotrainer.inference


__version__ = utilities.__version__

__all__ = [
    "datasets", "losses", "models", "optimizers", "solvers", "trainers", "utilities", "inference", "__version__"
]



from typing import Optional

from .fasta import read_FASTA

def convert_deprecated_fastas(result_file: str,
                              sequence_file: str,
                              labels_file: Optional[str] = None,
                              masks_file: Optional[str] = None,
                              skip_sequence_on_failed_merge: bool = False,
                              ):
    seqs = read_FASTA(sequence_file)
    labels = None
    masks = None

    if labels_file is not None:
        labels = {label_record.seq_id: label_record for label_record in read_FASTA(labels_file)}
    if masks_file is not None:
        masks = {mask_record.seq_id: mask_record for mask_record in read_FASTA(masks_file)}

    result_string = ""
    for seq_record in seqs:
        seq_id = seq_record.seq_id
        merged_target = seq_record.get_target()
        merged_set = seq_record.get_deprecated_set()
        merged_mask = seq_record.get_mask()
        # TARGET
        if labels and seq_id in labels:
            label_target = labels[seq_id].seq
            if merged_target is not None and merged_target != label_target:
                print(f"OVERWRITING sequence file TARGET for sequence {seq_id}!")
            merged_target = label_target

        # SET
        if labels and seq_id in labels:
            seq_set = seq_record.get_deprecated_set()
            labels_set = labels[seq_id].get_deprecated_set()
            if seq_set != labels_set and not (seq_set is not None or labels_set is not None):
                raise ValueError(f"Found ambiguity between sets for {seq_id} in sequences (SET={seq_set}) "
                                 f"and labels (SET={labels_set})!")
            else:
                merged_set = labels_set if labels_set is not None else seq_set

        # MASKS
        if masks and seq_id in masks:
            mask_file_mask = masks[seq_id].seq
            if merged_mask is not None and merged_mask != mask_file_mask:
                print(f"OVERWRITING mask file mask for sequence {seq_id}!")
            merged_mask = mask_file_mask

        # VALIDATE
        if merged_target is None:
            if skip_sequence_on_failed_merge:
                print(f"Could not merge target for seq_id {seq_id}, skipping sequence!")
                continue
            raise ValueError(f"Could not merge target for seq_id {seq_id}!")

        if labels_file and len(merged_target) != len(seq_record.seq):
            if skip_sequence_on_failed_merge:
                print(f"Sequence {seq_id} has incorrect target length, skipping sequence!")
                continue
            raise ValueError(f"Sequence {seq_id} has incorrect target length: "
                             f"{len(merged_target)} != {len(seq_record.seq)}!")

        if masks_file and merged_mask is not None and (len(merged_mask) != len(merged_target)):
            if skip_sequence_on_failed_merge:
                print(f"Sequence {seq_id} has incorrect mask length, skipping sequence!")
                continue
            raise ValueError(f"Sequence {seq_id} has incorrect mask length: "
                             f"{len(merged_mask)} != {len(merged_target)}!")

        if merged_set is None:
            if skip_sequence_on_failed_merge:
                print(f"Could not merge set for seq_id {seq_id}, skipping sequence!")
                continue
            raise ValueError(f"Could not merge set for seq_id {seq_id}!")

        merged_header = f">{seq_id} TARGET={merged_target} SET={merged_set}"
        if merged_mask:
            merged_header += f" MASK={merged_mask}"
        if seq_record.seq is None or seq_record.seq == "":
            raise ValueError(f"Could not find a valid sequence for seq_id {seq_id}!")
        result_string += merged_header + f"\n{seq_record.seq}\n"

    with open(result_file, "w") as f:
        f.write(result_string)



from __future__ import annotations
import numpy as np

from typing import Dict, Any, Union, Optional, List

from ..utilities import calculate_sequence_hash, RESIDUE_TO_VALUE_TARGET_DELIMITER


class BiotrainerSequenceRecord:
    def __init__(self, seq_id: str, seq: str, attributes: Optional[Dict[str, Any]] = None,
                 embedding: Optional[np.ndarray] = None):
        self.seq_id = seq_id
        self.seq = seq
        self.attributes = {k.upper(): v for k, v in attributes.items()} if attributes else {}
        self.embedding = embedding

    def get_target(self) -> Union[None, str, float]:
        return self.attributes.get("TARGET")

    def get_mask(self) -> Union[None, str]:
        return self.attributes.get("MASK")

    def get_set(self) -> Union[None, str]:
        return self.attributes.get("SET")

    def get_deprecated_set(self) -> Union[None, str]:
        if "SET" not in self.attributes:
            return None
        set_name = self.attributes["SET"]
        if set_name.lower() == "train":
            val = self.attributes.get("VALIDATION")
            if val is not None:
                val = eval(val)
                set_name = "val" if val else "train"
        return set_name

    def get_ppi(self) -> Union[None, str]:
        """ Get the INTERACTOR id (i.e. another sequence id in the same fasta file) """
        return self.attributes.get("INTERACTOR")

    def get_hash(self) -> str:
        return calculate_sequence_hash(self.seq)

    def copy_with_embedding(self, embedding: np.ndarray) -> BiotrainerSequenceRecord:
        """ Set the embedding for this sequence record and return sequence record """
        return BiotrainerSequenceRecord(seq_id=self.seq_id, seq=self.seq,
                                        attributes=self.attributes, embedding=embedding)

    @staticmethod
    def _convert_regression_target_if_necessary(target: str):
        if target is None:  # Can be the case for the predict dataset
            return target
        if RESIDUE_TO_VALUE_TARGET_DELIMITER in target:
            targets = target.split(RESIDUE_TO_VALUE_TARGET_DELIMITER)
            return list(map(float, targets))
        return target

    @staticmethod
    def get_dicts(input_records: List[BiotrainerSequenceRecord]) -> (dict, dict, dict):
        id2targets = {}
        id2masks = {}
        id2sets = {}
        for seq_record in input_records:
            seq_hash = seq_record.get_hash()

            target = seq_record.get_target()
            target = BiotrainerSequenceRecord._convert_regression_target_if_necessary(target)
            id2targets[seq_hash] = target

            mask = seq_record.get_mask()
            id2masks[seq_hash] = np.array([int(mask_value) for mask_value in mask]) if mask else None

            id2sets[seq_hash] = seq_record.get_set()
        return id2targets, id2masks, id2sets



from dataclasses import dataclass
from abc import ABC, abstractmethod
from typing import Optional, Dict, Any, Tuple

from ..utilities import EpochMetrics


@dataclass
class OutputData:
    config: Optional[Dict[str, Any]] = None
    derived_values: Optional[Dict[str, Any]] = None
    split_specific_values: Optional[Dict[str, Any]] = None
    training_iteration: Optional[Tuple[str, EpochMetrics]] = None
    test_results: Optional[Dict[str, Any]] = None
    predictions: Optional[Dict[str, Any]] = None


class BiotrainerOutputObserver(ABC):
    @abstractmethod
    def update(self, data: OutputData) -> None:
        """Handle an output event with associated data."""
        pass

    @abstractmethod
    def close(self) -> None:
        """Cleanup resources when training is finished."""
        pass


import torch
import numpy as np

from ..interfaces import EmbedderInterface

from ...utilities import AMINO_ACIDS

# Create a mapping of amino acids to their index
AA_TO_INDEX = {aa: i for i, aa in enumerate(AMINO_ACIDS)}


class OneHotEncodingEmbedder(EmbedderInterface):
    """
    Baseline embedder: One hot encoding as per-residue embedding, amino acid composition for per-protein

    This embedder is meant to be used as naive baseline for comparing different types of inputs or training methods.
    """

    embedding_dimension = len(AA_TO_INDEX.keys())
    name = "one_hot_encoding"

    def __init__(self):
        self.eye_matrix = np.eye(self.embedding_dimension, dtype=np.float32)

    def _embed_single(self, sequence: str) -> torch.tensor:
        # Convert sequence to indices
        indices = np.fromiter((AA_TO_INDEX.get(aa, -1) for aa in sequence), dtype=np.int8)

        # Use advanced indexing of identity matrix to create one-hot encoding
        return torch.tensor(self.eye_matrix[indices])

    @staticmethod
    def reduce_per_protein(embedding: torch.tensor) -> torch.tensor:
        """This returns the amino acid composition of the sequence as vector"""
        return embedding.mean(axis=0)



from typing import Dict, Any

from ..protocols import Protocol
from ..inference import Inferencer
from ..solvers import MetricsCalculator


class Bootstrapper:
    @staticmethod
    def bootstrap(protocol: Protocol, device, bootstrapping_iterations: int, metrics_calculator: MetricsCalculator,
                  mapped_predictions: Dict[str, Any], test_loader) -> Dict[str, Any]:
        try:
            max_prediction_length = len(max(mapped_predictions.values(), key=len))
        except TypeError:
            max_prediction_length = 1

        all_predictions_dict = {
            idx: Inferencer._pad_tensor(protocol=protocol,
                                        target=pred,
                                        length_to_pad=max_prediction_length,
                                        device=device) for idx, pred in mapped_predictions.items()
        }
        target_dict = {
            idx: Inferencer._pad_tensor(protocol=protocol,
                                        target=target,
                                        length_to_pad=max_prediction_length,
                                        device=device)
            for idx, target in
            zip(test_loader.dataset.ids, test_loader.dataset.targets)}
        seq_ids = list(target_dict.keys())

        sample_size = len(seq_ids)
        confidence_level = 0.05
        bootstrapping_results = Inferencer._do_bootstrapping(iterations=bootstrapping_iterations,
                                                             sample_size=sample_size,
                                                             confidence_level=confidence_level,
                                                             seq_ids=seq_ids,
                                                             all_predictions_dict=all_predictions_dict,
                                                             all_targets_dict=target_dict,
                                                             metrics_calculator=metrics_calculator.reset())
        bootstrapping_dict = {"results": bootstrapping_results, "iterations": bootstrapping_iterations,
                              "sample_size": sample_size, "confidence_level": confidence_level}
        return bootstrapping_dict



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created 2022

@author: Mason Minot
Original paper: https://www.biorxiv.org/content/10.1101/2022.03.08.483422v1
Original repository: https://github.com/minotm/NTA

@editor: Sebastian Franz
"""
import numpy as np
import itertools

from typing import Tuple, Dict, List

# dictionary of nucleotide codons and their corresponding amino acids
_nt_aa_dict = {
    'A': ['GCT', 'GCC', 'GCA', 'GCG'],
    'R': ['CGT', 'CGC', 'CGA', 'CGG', 'AGA', 'AGG'],
    'N': ['AAT', 'AAC'],
    'D': ['GAT', 'GAC'],
    'C': ['TGT', 'TGC'],
    'Q': ['CAA', 'CAG'],
    'E': ['GAA', 'GAG'],
    'G': ['GGT', 'GGC', 'GGA', 'GGG'],
    'H': ['CAT', 'CAC'],
    'I': ['ATT', 'ATC', 'ATA'],
    'L': ['CTT', 'CTC', 'CTA', 'CTG', 'TTA', 'TTG'],
    'K': ['AAA', 'AAG'],
    'M': ['ATG'],
    'F': ['TTT', 'TTC'],
    'P': ['CCT', 'CCC', 'CCA', 'CCG'],
    'S': ['TCT', 'TCC', 'TCA', 'TCG', 'AGT', 'AGC'],
    'T': ['ACT', 'ACC', 'ACA', 'ACG'],
    'W': ['TGG'],
    'Y': ['TAT', 'TAC'],
    'V': ['GTT', 'GTC', 'GTA', 'GTG'],
    '*': ['TAA', 'TGA', 'TAG'],
    'X': ['XXX']  # For masked input values
}


def _nucleotide_to_one_hot(nt: str) -> List:
    """
    @author: Sebastian Franz
    Parameters
    ----------
    nt: Single nucleotide (A,C,T,G,X)

    Returns
    -------
    List
        list of length 4 with one_hot_encoding of nucleotide
    """
    return {'A': [1, 0, 0, 0],
            'C': [0, 1, 0, 0],
            'T': [0, 0, 1, 0],
            'G': [0, 0, 0, 1],
            'X': [0, 0, 0, 0]}[nt]


def _nucleotide_trigram_to_one_hot(nt_trigram: str) -> np.ndarray:
    """
    @author: Sebastian Franz
    Parameters
    ----------
    nt_trigram: A nucleotide trigram, e.g. ACT or GGG

    Returns
    -------
    np.ndarray
        array of length 64 with one_hot_encoding of trigram
    """
    # Get all unique trigrams
    list_trigrams = list(set(itertools.chain.from_iterable(_nt_aa_dict.values())))
    list_trigrams.remove("XXX")  # Ignore masked value

    result = np.zeros(len(list_trigrams))
    if nt_trigram == "XXX":
        return result  # List of length 64 with only zeros, ignoring masked XXX value
    else:
        result[list_trigrams.index(nt_trigram)] = 1
    return result


def nucleotide_seq_to_unigram_vector(nt_seq: str) -> np.ndarray:
    """
    Flattens a nucleotide sequence to a vector of length 4 with the mean of the unigram encoded nucleotides

    @author Sebastian Franz
    Parameters
    ----------
    nt_seq: Sequence of nucleotides, e.g. AAATCTGGG

    Returns
    -------
    np.ndarray
        Vector of length 4, e.g. [0.5, 0.1, 0.4, 0.0]
    """
    result_list = []
    for nt in nt_seq:
        one_hot = _nucleotide_to_one_hot(nt)
        result_list.append(one_hot)
    return np.sum(result_list, axis=0) / len(result_list)


def nucleotide_seq_to_trigram_vector(nt_seq: str) -> np.ndarray:
    """
    Flattens a nucleotide sequence to a vector of length 64 with the mean of the trigram encoded nucleotides

    @author Sebastian Franz
    Parameters
    ----------
    nt_seq: str
        Sequence of nucleotides, e.g. AAATCTGGG

    Returns
    -------
    np.ndarray
        Vector of length 64, e.g. [0.5, 0.1, ..., 0.2, 0.0]
    """
    result_list = []
    trigram_length = 3
    chunks = [nt_seq[i:i + trigram_length] for i in range(0, len(nt_seq), trigram_length)]
    for nt_trigram in chunks:
        one_hot = _nucleotide_trigram_to_one_hot(nt_trigram)
        result_list.append(one_hot)
    return np.sum(result_list, axis=0) / len(result_list)


def aa_to_nt(seqs_and_targets_dict: dict, aug_factor: int) -> Tuple[Dict, Dict]:
    """
    Function which produces aug_factor unique nucleotide sequences for each amino acid sequence in
    seq_list. The appropriate targets (labels) are maintained for the augmented sequences.

    @author: Mason Minot
    @editor: Sebastian Franz
    Adapted to the needs of biotrainer. Input is replaced by one dictionary with seq_id -> (sequence, target).
    Thereby, results are directly stored by original sequence id, making things easier
    in the notebook.

    Parameters
    ----------
    seqs_and_targets_dict: dict
        seq_id: str -> (sequence: str, target: str)

    aug_factor : int
        the augmentation factor. the number of unique nucleotide sequences to create per protein sequence

    Returns
    -------
    seq_dict : dict
        seq_id -> List with nucleotide sequences (length of list == aug_factor)
    target_dict:
        seq_id -> List of targets (length of list == aug_factor, not necessary for the notebook)
    """

    seq_by_id = {idx: str(sequence) for idx, (sequence, target) in seqs_and_targets_dict.items()}
    target_by_id = {idx: target for idx, (sequence, target) in seqs_and_targets_dict.items()}

    seq_dict = {}
    target_dict = {}
    for key in seqs_and_targets_dict.keys():
        seq_dict[key] = []
        nt_codons_per_residue = {}
        for i in range(len(seq_by_id[key])):
            # determine possible nt codons per aa position
            nt_codons_per_residue[str(i)] = _nt_aa_dict[seq_by_id[key][i]]

        # use itertools product function to create a list of all possible combinations of nt codons  for given aa seq
        nucleotides = list(itertools.islice(itertools.product(*nt_codons_per_residue.values()), aug_factor))
        # convert list of tuples to list of strings
        nucleotides = list(map(''.join, nucleotides))
        tmp_target_list = []
        for j in range(len(nucleotides)):
            tmp_target_list.append(target_by_id[key])
        seq_dict[key] = nucleotides
        target_dict[key] = tmp_target_list

    return seq_dict, target_dict


"""
# Commented out because not necessary for the notebook 
def nt_augmentation(input_seqs, final_data_len=2e5, is_val_set=False):
    '''
    Wrapper function to setup nucleotide augmentation based on a desired augmented data length. If
    is_val_set = True, then sequences will be backtranslated (from amino acids to nucleotides) without
    augmentation
    
    @author Mason Minot
    Parameters
    ----------
    input_seqs : list or pandas.core.series.Series
        list or series of amino acid sequences
    final_data_len : int
        desired length of final data set
    is_val_set : bool
        whether or not input_seqs is a validation set. If is_val_set = True, backtranslation without
        augmentation is performed.

    Returns
    -------
    out_df : pandas.core.frame.DataFrame
        pandas dataframe containing augmented nucleotide sequences
    '''
    data_len = len(input_seqs['aaseq'])

    # round the calculated fraction to a whole number to get qty for first augmentation
    # this will augment the data greater than is necessary. i.e. for desired aug of 1.5, it will augment 2x
    calculated_aug_factor = int(ceil(final_data_len / data_len))

    if calculated_aug_factor == 0 or is_val_set == True:
        calculated_aug_factor = 1
    # for her2 negatives, augmenting by 2 then subsampling will decrease the sequence diveristy of the majority class
    # essentially acting as a majority class downsampling effect
    # elif is_val_set =='her2_neg': calculated_aug_factor = 2

    data = input_seqs.copy()
    aa_seq_list = data['aaseq']
    target_list = data['target']
    seq_dict, target_dict = aa_to_nt(aa_seq_list, target_list=target_list, aug_factor=calculated_aug_factor)

    # randomly downsample augmented data set to desired length
    if is_val_set == False:
        truncate_factor = final_data_len / data_len
        len_seq_dict = sum([len(x) for x in seq_dict.values()])  # number of total nucleotide sequences in dictionary

        # downsample augmented sequences by iteratively dropping one augmented nt seq from each
        # aa seq until desired data size is reached
        if final_data_len < len_seq_dict:
            num_seqs_to_drop = int(len_seq_dict - final_data_len)
            for i in range(num_seqs_to_drop):
                seq_dict[i] = np.random.choice(seq_dict[i], len(seq_dict[i]) - 1, replace=False)
                target_dict[i] = np.random.choice(target_dict[i], len(target_dict[i]) - 1, replace=False)

    seq_out_list = []
    target_out_list = []
    for key in seq_dict:
        for seq_entry in seq_dict[key]:
            seq_out_list.append(seq_entry)
        for target_entry in target_dict[key]:
            target_out_list.append(target_entry)

    out_df = pd.DataFrame(seq_out_list)
    out_df.columns = ['dnaseq']
    out_df['target'] = target_out_list

    return out_df
"""


from dataclasses import dataclass
from typing import Dict, Any, Optional


@dataclass
class AutoEvalProgress:
    completed_tasks: int
    total_tasks: int
    current_framework_name: str
    current_task_name: str
    final_report: Optional[Dict[str, Any]] = None



from pathlib import Path
from datasets import load_dataset
from typing import Any, Dict, List, Optional, Tuple

from ..protocols import Protocol
from ..utilities import get_logger

logger = get_logger(__name__)


def _write_hf_to_fasta(
        sequences: List[str],
        targets: List[Any],
        set_values: List[str],
        hf_storage_path: Path,
        masks: Optional[List[Any]] = None, ) -> Optional[Path]:
    """
    Converts sequences, targets, and optional masks from a HuggingFace dataset into FASTA file(s).

    Args:
        sequences (List[str]): A list of protein sequences.
        targets (List[Any]): A list of target values.
        set_values (List[str]): A list of SET values corresponding to each sequence.
        hf_storage_path (Path): Path to the HuggingFace storage path.
        masks (Optional[List[Any]]): A list of mask values, or None if masks are not provided.

    Raises:
        ValueError: If the lengths of sequences, targets, masks (if masks are provided), and set_values do not match.
        IOError: If there is an issue writing to the output files.
    """
    if not masks:
        if not (len(sequences) == len(targets) == len(set_values)):
            raise ValueError("The number of sequences, targets, and set_values must be the same.")
    else:
        if not (len(sequences) == len(targets) == len(set_values) == len(masks)):
            raise ValueError("The number of sequences, targets, set_values, and masks must be the same.")

    input_file_name = hf_storage_path / "input.fasta"
    with open(input_file_name, "w") as input_file:
        # Write the data
        for idx in range(len(sequences)):
            seq_id = f"Seq{idx + 1}"
            seq = sequences[idx]
            target = targets[idx]
            set_val = set_values[idx]
            mask = masks[idx] if masks is not None else None

            seq_header = f">{seq_id} TARGET={target} SET={set_val}"
            if mask is not None:
                seq_header += f" MASK={mask}"
            input_file.write(f"{seq_header}\n{seq}\n")

        return input_file_name


def _process_subset(
        current_subset: Any,
        sequence_column: str,
        target_column: str,
        mask_column: Optional[str] = None
) -> Tuple[List[str], List[Any], Optional[List[Any]]]:
    """
    Processes a single subset, verifying the presence of required columns
    and extracting sequences, targets, and optionally masks.

    Args:
        current_subset (Any): The subset to process. This is typically a HuggingFace dataset or similar structure.
        sequence_column (str): The name of the column containing the sequences.
        target_column (str): The name of the column containing the target values.
        mask_column (Optional[str]): The name of the column containing mask values, if applicable. Defaults to None.

    Returns:
        Tuple[List[str], List[Any], Optional[List[Any]]]: A tuple containing:
            - List[str]: A list of sequences.
            - List[Any]: A list of target values.
            - Optional[List[Any]]: A list of mask values, or None if no mask column is provided.

    Raises:
        Exception: If any of the specified columns are missing from the dataset.
    """
    # Verify columns
    _verify_dataset_column(current_subset, sequence_column)
    _verify_dataset_column(current_subset, target_column)
    if mask_column:
        _verify_dataset_column(current_subset, mask_column)

    # Extract data
    sequences = current_subset[sequence_column]
    targets = current_subset[target_column]
    masks = current_subset[mask_column] if mask_column else None

    return sequences, targets, masks


def _determine_set_name(subset_name: str) -> str:
    """
    Determines the corresponding set name ("TRAIN", "VAL", "TEST", "PRED") based on the provided subset name.

    This function normalizes and categorizes the input subset name into one of the standard set names. If the input does not
    match any of the expected patterns ("train", "val", "test", "pred"),
    it logs a warning and assigns the subset to "TEST-{subset_name}" by default.

    Args:
        subset_name (str): The name of the subset (e.g., "train1", "validation", "testing").

    Returns:
        str: The normalized set name ("TRAIN", "VAL", "TEST", "PRED", "TEST-{subset_name}").

    Logs:
        Warning: Logs a warning if the subset name is unrecognized and defaults to "TEST".

    """
    lower_subset_name = subset_name.lower()
    if lower_subset_name.startswith("train"):
        return "TRAIN"
    elif lower_subset_name.startswith("val"):
        return "VAL"
    elif lower_subset_name.startswith("test"):
        return "TEST"
    elif lower_subset_name.startswith("pred"):
        return "PRED"
    else:
        logger.warning(f"Unrecognized subset name '{subset_name}'. Assigning to 'TEST-{lower_subset_name}'.")
        return f"TEST-{lower_subset_name}"


def _verify_dataset_column(dataset: Any, column: str) -> None:
    """
    Verifies that the specified column exists in the given dataset.

    Args:
        dataset (Any): The dataset to verify. This is expected to have a `column_names` attribute,
                       such as a HuggingFace dataset or a similar structure.
        column (str): The name of the column to check for existence.

    Raises:
        Exception: If the specified column is not found in the dataset.
    """
    if column not in dataset.column_names:
        raise Exception(
            f"Column '{column}' not found in the dataset."
        )


def _load_and_split_hf_dataset(hf_map: Dict) -> Tuple[List[str], List[Any], List[Any], List[str]]:
    """
    Loads a HuggingFace dataset and splits it into sequences, targets, masks (if available), and set values.

    Args:
        hf_map (Dict[str, Any]): A mapping of configuration options for loading the HuggingFace dataset.
                                 Expected keys include:
                                 - "path" (str): The dataset path.
                                 - "sequence_column" (str): Name of the sequence column.
                                 - "target_column" (str): Name of the target column.
                                 - "mask_column" (Optional[str]): Name of the mask column (if applicable).
                                 - "subset" (Optional[str]): Subset name, if required.

    Returns:
        Tuple[List[str], List[Any], Optional[List[Any]], List[str]]:
            - List[str]: Sequences extracted from the dataset.
            - List[Any]: targets corresponding to the sequences.
            - Optional[List[Any]]: Masks, if available; otherwise, None.
            - List[str]: Set values (e.g., "TRAIN", "VAL", "TEST") for each sequence.

    Raises:
        ValueError: If loading the dataset fails due to missing or invalid configuration.
        Exception: If the required splits or columns are missing, or if other errors occur during processing.

    """
    # Extract parameters from hf_map
    path = hf_map["path"]
    sequence_column = hf_map["sequence_column"]
    target_column = hf_map["target_column"]
    mask_column = hf_map["mask_column"] if "mask_column" in hf_map else None
    subset_name = hf_map["subset"] if "subset" in hf_map else None

    logger.info(f"Loading HuggingFace dataset from path: {path}")

    # Load dataset
    try:
        dataset = load_dataset(path, subset_name if subset_name is not None else 'default')
    except ValueError as e:
        error_msg = ("Loading the dataset from Hugging Face failed. "
                     "If the dataset requires a 'subset', you can specify it using the 'subset' option in the config file.\n"
                     f"Error: {e}")
        raise ValueError(error_msg)
    except Exception as e:
        raise Exception(f"Loading the dataset from Hugging Face failed. Error: {e}")

    # Collect all available subsets
    subset_names = list(dataset.keys())
    if not subset_names:
        raise Exception(f"No subsets found in the dataset at path '{path}'.")

    sequences = []
    targets = []
    masks = []
    set_values = []

    if len(subset_names) >= 3:
        logger.info(f"{len(subset_names)} subsets found: {subset_names}. Processing each separately.")
        for subset_name in subset_names:
            current_subset = dataset[subset_name]
            if current_subset is None:
                logger.warning(f"Subset '{subset_name}' is None. Skipping.")
                continue

            # Process subset
            current_sequences, current_targets, current_masks = _process_subset(current_subset, sequence_column,
                                                                                target_column, mask_column)

            # Determine SET value based on split name
            current_set_value = _determine_set_name(subset_name)

            sequences.extend(current_sequences)
            targets.extend(current_targets)
            set_values.extend([current_set_value] * len(current_sequences))
            if current_masks is not None:
                masks.extend(current_masks)
            else:
                masks.extend([None] * len(current_sequences))

    else:
        raise ValueError(
            f"Expected at least 3 subsets (TRAIN, VAL, TEST) in the dataset at path '{path}'. Found: {subset_names}."
        )

    return sequences, targets, masks, set_values


def process_hf_dataset_to_fasta(
        hf_storage_path: Path,
        hf_map: Dict) -> Optional[Path]:
    """
    Loads a HuggingFace dataset, splits it according to the protocol, and writes the data to FASTA files.

    Args:
        hf_storage_path (Path): Path to the HuggingFace storage directory.
        hf_map (Dict): A mapping of HuggingFace dataset option names to their respective ConfigOption instances.

    Raises:
        Exception: If there is an issue during the creation of the required files or processing the dataset.
    """
    try:
        sequences, targets, masks, set_values = _load_and_split_hf_dataset(hf_map)
    except Exception as e:
        raise Exception(f"Failed to load and split HuggingFace dataset: {e}")

    try:
        input_file_path = _write_hf_to_fasta(
            sequences=sequences,
            targets=targets,
            masks=masks if any([m is not None for m in masks]) else None,
            set_values=set_values,
            hf_storage_path=hf_storage_path,
        )

        logger.info("HuggingFace dataset downloaded and processed successfully.")
        return input_file_path
    except Exception as e:
        raise Exception(f"Failed to write huggingface dataset to FASTA files: {e}")



import torch
import torch.nn as nn

from typing import Optional, Union, Dict, Any, Set

from .masked_mse_loss import MaskedMSELoss

from ..protocols import Protocol
from ..utilities import MASK_AND_LABELS_PAD_VALUE

__LOSSES = {
    Protocol.residue_to_class: {
        'cross_entropy_loss': lambda **kwargs: nn.CrossEntropyLoss(**kwargs, ignore_index=MASK_AND_LABELS_PAD_VALUE)
    },
    Protocol.residue_to_value: {
        'mean_squared_error': lambda **kwargs: MaskedMSELoss(ignore_index=MASK_AND_LABELS_PAD_VALUE)
    },
    Protocol.residues_to_class: {
        'cross_entropy_loss': lambda **kwargs: nn.CrossEntropyLoss(**kwargs, ignore_index=MASK_AND_LABELS_PAD_VALUE)
    },
    Protocol.residues_to_value: {
        'mean_squared_error': lambda **kwargs: nn.MSELoss(**kwargs)
    },
    Protocol.sequence_to_class: {
        'cross_entropy_loss': lambda **kwargs: nn.CrossEntropyLoss(**kwargs, ignore_index=MASK_AND_LABELS_PAD_VALUE)
    },
    Protocol.sequence_to_value: {
        'mean_squared_error': lambda **kwargs: nn.MSELoss(**kwargs)
    },
}


def get_loss(protocol: Protocol, loss_choice: str, device: Union[str, torch.device],
             weight: Optional[torch.Tensor] = None, **kwargs):
    loss = __LOSSES.get(protocol).get(loss_choice)

    if not loss:
        raise NotImplementedError
    else:
        if weight is not None:
            return loss(weight=weight).to(device)
        else:
            return loss().to(device)


def get_available_losses_dict() -> Dict[Protocol, Dict[str, Any]]:
    return dict(__LOSSES)


def get_available_losses_set() -> Set[str]:
    all_losses_by_protocol = [list(protocol.keys()) for protocol in get_available_losses_dict().values()]
    all_losses_list = [loss for loss_list in all_losses_by_protocol for loss in loss_list]
    return set(all_losses_list)


__all__ = [
    "get_loss",
    "get_available_losses_dict",
    "get_available_losses_set"
]



from .training_factory import TrainingFactory

from ..pipeline import PipelineContext, PipelineStep
from ..pipeline.pipeline_step import PipelineStepType

from ...solvers import get_metrics_calculator
from ...utilities import get_logger, revert_mappings
from ...validations import SanityChecker, Bootstrapper

logger = get_logger(__name__)


class TestingStep(PipelineStep):

    def get_step_type(self) -> PipelineStepType:
        return PipelineStepType.TESTING

    @staticmethod
    def _do_and_log_evaluation(context: PipelineContext, solver, test_loader, test_set_id: str):
        # re-initialize the model to avoid any undesired information leakage and only load checkpoint weights
        solver.load_checkpoint(resume_training=False)
        test_results = solver.inference(test_loader, calculate_test_metrics=True)

        if context.config.get("save_split_ids", False):
            test_results['test_set_predictions'] = revert_mappings(protocol=context.config["protocol"],
                                                                   test_predictions=test_results['mapped_predictions'],
                                                                   class_int2str=context.target_manager.class_int2str)
            context.output_manager.add_test_set_result(test_set_id=test_set_id,
                                                       test_set_results={k: v for k, v in test_results.items()
                                                                         if k != "mapped_probabilities"})
        else:
            context.output_manager.add_test_set_result(test_set_id=test_set_id,
                                                       test_set_results={'metrics': test_results['metrics']})

        logger.info(f"Test set {test_set_id} metrics: {test_results['metrics']}")
        return test_results

    @staticmethod
    def _do_and_log_prediction(context: PipelineContext, solver, pred_loader):
        # re-initialize the model to avoid any undesired information leakage and only load checkpoint weights
        solver.load_checkpoint(resume_training=False)
        pred_results = solver.inference(pred_loader, calculate_test_metrics=False)

        predictions = revert_mappings(protocol=context.config["protocol"],
                                      test_predictions=pred_results['mapped_predictions'],
                                      class_int2str=context.target_manager.class_int2str)
        context.output_manager.add_prediction_result(prediction_results=predictions)

        logger.info(f"Calculated predictions for {len(pred_loader)} samples!")
        return pred_results

    @staticmethod
    def _do_and_log_bootstrapping_evaluation(context: PipelineContext,
                                             metrics_calculator,
                                             test_results, test_loader, test_set_id: str):
        logger.info(f'Running bootstrapping evaluation on the best model for test set ({test_set_id})')
        bootstrapping_dict = Bootstrapper.bootstrap(protocol=context.config["protocol"],
                                                    device=context.config["device"],
                                                    bootstrapping_iterations=context.config["bootstrapping_iterations"],
                                                    metrics_calculator=metrics_calculator,
                                                    mapped_predictions=test_results["mapped_predictions"],
                                                    test_loader=test_loader)
        context.output_manager.add_test_set_result(test_set_id=test_set_id,
                                                   test_set_results={"bootstrapping": bootstrapping_dict})
        logger.info(f'Bootstrapping results for test set ({test_set_id}): {bootstrapping_dict}')

    def process(self, context: PipelineContext) -> PipelineContext:
        # TESTING
        test_datasets = context.test_datasets
        best_split = context.best_split
        assert test_datasets is not None, f"test_datasets cannot be None at the testing step!"
        assert best_split is not None, f"best_split cannot be None at the testing step!"

        finetuning = "finetuning_config" in context.config
        for test_set_id, test_dataset in test_datasets.items():
            logger.info('Running final evaluation on the best model')
            test_dataset_embeddings = TrainingFactory.create_dataset(context=context,
                                                                     split=test_dataset,
                                                                     mode="test",
                                                                     finetuning=finetuning)
            test_loader = TrainingFactory.create_dataloader(context=context, dataset=test_dataset_embeddings,
                                                            hyper_params=best_split.hyper_params, finetuning=finetuning)
            test_results = self._do_and_log_evaluation(context=context,
                                                       solver=best_split.solver,
                                                       test_loader=test_loader,
                                                       test_set_id=test_set_id)

            # ADDITIONAL EVALUATION
            metrics_calculator = get_metrics_calculator(protocol=context.config["protocol"],
                                                        device=context.config["device"],
                                                        n_classes=context.n_classes)
            # BOOTSTRAPPING
            if context.config["bootstrapping_iterations"] > 0:
                self._do_and_log_bootstrapping_evaluation(context=context,
                                                          metrics_calculator=metrics_calculator,
                                                          test_results=test_results,
                                                          test_loader=test_loader,
                                                          test_set_id=test_set_id)

            # SANITY CHECKER
            if context.config.get("sanity_check", True):
                baseline_test_dataset = context.baseline_test_datasets[test_set_id]
                baseline_test_dataset_embeddings = TrainingFactory.create_dataset(context=context,
                                                                                  split=baseline_test_dataset,
                                                                                  mode="test",
                                                                                  finetuning=False)
                baseline_test_loader = TrainingFactory.create_dataloader(context=context,
                                                                         dataset=baseline_test_dataset_embeddings,
                                                                         hyper_params=best_split.hyper_params,
                                                                         finetuning=False)
                sanity_checker = SanityChecker(training_config=context.config,
                                               n_classes=context.n_classes,
                                               n_features=context.n_features,
                                               train_dataset=context.train_dataset,
                                               val_dataset=context.val_dataset,
                                               test_dataset=baseline_test_dataset,
                                               test_loader=baseline_test_loader,
                                               metrics_calculator=metrics_calculator,
                                               test_results_dict=test_results,
                                               mode="warn")
                baseline_results, warnings = sanity_checker.check_test_results(test_set_id=test_set_id)
                if baseline_results is not None and len(baseline_results) > 0:
                    context.output_manager.add_test_set_result(test_set_id=test_set_id,
                                                               test_set_results={"test_baselines": baseline_results})
                if len(warnings) > 0:
                    context.output_manager.add_test_set_result(test_set_id=test_set_id,
                                                               test_set_results={"sanity_check_warnings": warnings})

        # PREDICTION
        prediction_dataset = context.prediction_dataset
        if prediction_dataset and len(prediction_dataset) > 0:
            logger.info(f'Calculating predictions for {len(prediction_dataset)} samples!')
            pred_dataset_embeddings = TrainingFactory.create_dataset(context=context,
                                                                     split=prediction_dataset, mode="pred")
            pred_loader = TrainingFactory.create_dataloader(context=context, dataset=pred_dataset_embeddings,
                                                            hyper_params=best_split.hyper_params, finetuning=finetuning)

            _ = self._do_and_log_prediction(context=context,
                                            solver=best_split.solver,
                                            pred_loader=pred_loader
                                            )

        return context



import re

from pathlib import Path
from typing import Union, List, Callable

from .biotrainer_seq_record import BiotrainerSequenceRecord


def read_FASTA(path: Union[str, Path]) -> List[BiotrainerSequenceRecord]:
    """
    Pure Python FASTA file parser. Parses attributes on the go.

    :param path: path to a valid FASTA file
    :return: a list of BiotrainerSequenceRecord objects
    """
    attributes_pattern = re.compile(r"([A-Z_]+)=([^ ]+)")  # Allow any combination of ATTRIBUTE_NAME=VALUE

    records = []
    try:
        with open(path, 'r') as file:
            current_id = ""
            current_attributes = {}
            current_seq = ""

            for line in file:
                line = line.strip()
                if not line:
                    continue

                if line.startswith('>'):
                    # Save the previous sequence if it exists
                    if current_id:
                        records.append(BiotrainerSequenceRecord(
                            seq_id=current_id,
                            attributes=current_attributes,
                            seq=current_seq
                        ))

                    # Parse the header line
                    header = line[1:].strip()
                    parts = header.split(maxsplit=1)
                    current_id = parts[0]
                    current_description = parts[1] if len(parts) > 1 else ""
                    current_attributes = {key: value for key, value in attributes_pattern.findall(current_description)}
                    current_seq = ""
                else:
                    current_seq = line

            # Ensure last sequence is added
            if current_id:
                records.append(BiotrainerSequenceRecord(
                    seq_id=current_id,
                    attributes=current_attributes,
                    seq=current_seq
                ))

        return records

    except FileNotFoundError as e:
        raise e
    except ValueError as e:
        raise e
    except Exception as e:
        raise ValueError(f"Could not parse '{path}'. Are you sure this is a valid fasta file?") from e


def write_FASTA(path: Union[str, Path], seq_records: List[BiotrainerSequenceRecord]) -> int:
    """
    Save a list of BiotrainerSequenceRecord objects to a FASTA file.

    :param path: Path to store the file
    :param seq_records: List of BiotrainerSequenceRecord objects
    :return: number of sequence records written (following biopython standard)
    """
    with open(path, 'w') as fasta_file:
        n_written = 0
        for seq_record in seq_records:
            attributes_str = " ".join([f"{key.upper()}={value}" for key, value in seq_record.attributes.items()])
            fasta_file.write(f">{seq_record.seq_id} {attributes_str}\n{seq_record.seq}\n")
            n_written += 1
    return n_written


def filter_FASTA(input_path: Union[str, Path], output_path: Union[str, Path],
                 filter_function: Callable[[BiotrainerSequenceRecord], bool]) -> (int, int):
    """
    Reads a fasta file from input_path, applies the filter function to all seq_records
    and stores the result at output_path.

    :param input_path: The path to a fasta file
    :param output_path: The path were the filtered fasta file will be written
    :param filter_function: Callable that takes a BiotrainerSequenceRecord and returns a bool (True == keep)
    :return: (Number of kept sequences, Number of all sequences)
    """
    seq_records = read_FASTA(input_path)
    n_all = len(seq_records)
    keep_records = [seq_record for seq_record in seq_records if filter_function(seq_record)]
    n_kept = write_FASTA(output_path, keep_records)
    return n_kept, n_all


from typing import Tuple, List

from .config_option import ConfigOption, ConfigConstraints, ConfigKey

from ..protocols import Protocol

def input_config(protocol: Protocol) -> Tuple[ConfigKey, List[ConfigOption]]:
    input_category = "input_files"
    return ConfigKey.ROOT, [
        ConfigOption(
            name="input_file",
            description="Provide the input file, which contains the biological sequences, "
                        "targets and optionally masks to be used for training.",
            category=input_category,
            required=False,  # Can be replaced by hf_dataset config
            is_file_option=True,
            constraints=ConfigConstraints(
                type=str,
                allowed_formats=[".fasta"]
            )
        ),
    ]


from pathlib import Path
from typing import Dict, Any, List, Union

from .tensorboard_writer import TensorboardWriter
from .biotrainer_output_observer import BiotrainerOutputObserver
from .output_manager import OutputManager, InferenceOutputManager

def output_observer_factory(output_dir: Union[Path, str], config: Dict[str, Any]) -> List[BiotrainerOutputObserver]:
    result = []
    if "external_writer" in config:
        external_writer: str = config["external_writer"]
        if external_writer.lower() == "none":
            return []
        if external_writer.lower() == "tensorboard":
          result.append(TensorboardWriter(log_dir=Path(output_dir) / "runs"))
    return result

__all__ = ["OutputManager", "InferenceOutputManager", "BiotrainerOutputObserver", "output_observer_factory"]



import torch
import numpy as np
import blosum as bl

from ..interfaces import EmbedderInterface

from ...utilities import AMINO_ACIDS


class Blosum62Embedder(EmbedderInterface):
    """
    Baseline embedder: BLOSUM62 matrix as per-residue embedding

    Each amino acid is represented by its corresponding row in the BLOSUM62 substitution matrix.
    This provides evolutionary information about amino acid substitution patterns.
    """

    name = "blosum62"

    def __init__(self):
        # Create efficient lookup structures
        self._setup_efficient_mapping()

    def _setup_efficient_mapping(self):
        """Setup efficient amino acid to BLOSUM score mapping"""
        # Load BLOSUM62 matrix
        matrix = bl.BLOSUM(62)

        self.embedding_dimension = len(AMINO_ACIDS)  # 21 amino acids including X

        # Create mapping from amino acid to index
        self.aa_to_index = {aa: i for i, aa in enumerate(AMINO_ACIDS)}

        # Create the lookup matrix: shape (21, 21) for all AAs including X
        self.lookup_matrix = np.zeros((len(AMINO_ACIDS), self.embedding_dimension), dtype=np.float32)

        # Fill the lookup matrix with BLOSUM values
        for i, aa in enumerate(AMINO_ACIDS):
            # Get the BLOSUM row for this amino acid in the correct order
            blosum_row = [matrix[aa][target_aa] for target_aa in AMINO_ACIDS]
            self.lookup_matrix[i] = np.array(blosum_row, dtype=np.float32)

    def _embed_single(self, sequence: str) -> torch.tensor:
        """Convert sequence to BLOSUM-based embedding"""
        indices = np.fromiter(
            (self.aa_to_index.get(aa, self.aa_to_index['X']) for aa in sequence),
            dtype=np.int32
        )

        # Use advanced indexing to get BLOSUM rows for each position
        return torch.tensor(self.lookup_matrix[indices])

    @staticmethod
    def reduce_per_protein(embedding: torch.tensor) -> torch.tensor:
        """Returns the average BLOSUM profile of the sequence"""
        return embedding.mean(axis=0)



import tempfile
import unittest
import numpy as np
import torch._dynamo

from sys import platform
from typing import Dict, Union, List

from biotrainer.inference import Inferencer
from biotrainer.embedders import OneHotEncodingEmbedder
from biotrainer.protocols import Protocol


class InferencerTests(unittest.TestCase):
    _test_sequences = ["SEQVENCEPRTEI",
                       "SEQWENCE",
                       "PRTEIN"]
    _test_targets_r2c = ["CDEVVCCCDDDDD",
                         "DDDDDDDD",
                         "CDEVVV"]
    _test_targets_r2v = [[0.05, 1.0, 2, 1.1, -1.2, 1.11, 5.4, 33, 0.12, -1.0, 0.0, 12, 14.131],
                         [1, 2, 3, 4, 5, 6, 7, 8],
                         [-1, 0, -0, 5.5, 1.01, 12.0001]]
    _test_targets_rs2c = ["TM", "TMSP", "Glob"]
    _test_targets_s2c = ["TM", "TMSP", "Glob"]
    _test_classes_r2c = ["C", "D", "E", "F", "V"]
    _test_classes_rs2c = ["Glob", "GlobSP", "TM", "TMSP"]
    _test_classes_s2c = ["Glob", "GlobSP", "TM", "TMSP"]
    _test_targets_s2v = [1, -1.212, 0.0]
    _test_targets_rs2v = [5, -1.212, 0.0]

    error_tolerance = 0.02
    error_tolerance_factor = 1.96

    def setUp(self) -> None:
        # Needed for cross-platform compatibility of Windows and Linux models
        torch._dynamo.config.suppress_errors = True

        self.inferencer_r2c, _ = Inferencer.create_from_out_file("test_input_files/test_models/r2c/out.yml")
        self.inferencer_r2v, _ = Inferencer.create_from_out_file("test_input_files/test_models/r2v/out.yml")
        self.inferencer_rs2c, _ = Inferencer.create_from_out_file("test_input_files/test_models/rs2c/out.yml")
        self.inferencer_s2c, _ = Inferencer.create_from_out_file("test_input_files/test_models/s2c/out.yml")
        self.inferencer_s2v, _ = Inferencer.create_from_out_file("test_input_files/test_models/s2v/out.yml")
        self.inferencer_rs2v, _ = Inferencer.create_from_out_file("test_input_files/test_models/rs2v/out.yml")

        self.inferencer_list = [self.inferencer_r2c, self.inferencer_rs2c, self.inferencer_s2c, self.inferencer_s2v,
                                self.inferencer_rs2v]

        self.per_residue_embeddings, self.per_sequence_embeddings = self._embed()

    @classmethod
    def tearDownClass(cls) -> None:
        # Undo error suppressing in dynamo
        torch._dynamo.config.suppress_errors = False

    def _embed(self):
        embedder = OneHotEncodingEmbedder()
        per_residue_embeddings = list(embedder.embed_many(self._test_sequences))
        per_sequence_embeddings = [embedder.reduce_per_protein(embedding) for embedding in
                                   per_residue_embeddings]
        per_residue_embeddings_dict = {f"Seq{idx}": embed for idx, embed in enumerate(per_residue_embeddings)}
        per_sequence_embeddings_dict = {f"Seq{idx}": embed for idx, embed in enumerate(per_sequence_embeddings)}
        return per_residue_embeddings_dict, per_sequence_embeddings_dict

    def test_padding(self):
        max_seq_length = len(max(self._test_targets_r2c, key=len))
        for target in self._test_targets_r2c:
            self.assertTrue(len(self.inferencer_r2c._pad_tensor(protocol=self.inferencer_r2c.protocol,
                                                                target=self.inferencer_r2c._convert_class_str2int(
                                                                    target),
                                                                length_to_pad=max_seq_length,
                                                                device="cpu")) == max_seq_length,
                            "Padding did not enlarge target sequence to correct size!")
        for target in self._test_targets_s2v:
            self.assertTrue(self.inferencer_s2v._pad_tensor(protocol=self.inferencer_s2v.protocol, target=target,
                                                            length_to_pad=max_seq_length, device="cpu") == target,
                            "Padding changed a non-list value!")

    def test_from_embeddings(self):
        r2c_dict = self.inferencer_r2c.from_embeddings(self.per_residue_embeddings)
        r2v_dict = self.inferencer_r2v.from_embeddings(self.per_residue_embeddings)
        rs2c_dict = self.inferencer_rs2c.from_embeddings(self.per_residue_embeddings)
        s2c_dict = self.inferencer_s2c.from_embeddings(self.per_sequence_embeddings)
        s2v_dict = self.inferencer_s2v.from_embeddings(self.per_sequence_embeddings)
        rs2v_dict = self.inferencer_rs2v.from_embeddings(self.per_residue_embeddings)

        self.assertTrue(
            "metrics" in r2c_dict.keys() and "metrics" in r2v_dict.keys() and
            "metrics" in s2c_dict.keys() and "metrics" in s2v_dict.keys(),
            "Missing metrics key!")
        self.assertTrue("mapped_predictions" in r2c_dict.keys() and "mapped_predictions" in s2v_dict.keys(),
                        "Missing predictions key!")

        self.assertTrue(all([pred_class in self._test_classes_r2c for pred_class in
                             "".join(r2c_dict["mapped_predictions"].values())]),
                        "Inferencer predicted a non-existing class!")
        self.assertTrue(all([type(value) for value in
                             [v for v in (vs for vs in r2v_dict["mapped_predictions"].values())]]),
                        "Inferencer predicted a non-existing class!")
        self.assertTrue(all([pred_class in self._test_classes_rs2c
                             for pred_class in rs2c_dict["mapped_predictions"].values()]),
                        "Inferencer predicted a non-existing class (rs2c)!")
        self.assertTrue(all([pred_class in self._test_classes_s2c
                             for pred_class in s2c_dict["mapped_predictions"].values()]),
                        "Inferencer predicted a non-existing class (s2c)!")
        self.assertTrue(all([type(value) is float for value in s2v_dict["mapped_predictions"].values()]),
                        "Type of all sequence to value predictions is not float!")
        self.assertTrue(all([type(value) is float for value in rs2v_dict["mapped_predictions"].values()]),
                        "Type of all residues to value predictions is not float!")

    def test_from_embeddings_with_targets(self):
        r2c_dict = self.inferencer_r2c.from_embeddings(self.per_residue_embeddings, self._test_targets_r2c)
        r2v_dict = self.inferencer_r2v.from_embeddings(self.per_residue_embeddings, self._test_targets_r2v)
        rs2c_dict = self.inferencer_rs2c.from_embeddings(self.per_residue_embeddings, self._test_targets_rs2c)
        s2c_dict = self.inferencer_s2c.from_embeddings(self.per_sequence_embeddings, self._test_targets_s2c)
        s2v_dict = self.inferencer_s2v.from_embeddings(self.per_sequence_embeddings, self._test_targets_s2v)
        rs2v_dict = self.inferencer_rs2v.from_embeddings(self.per_residue_embeddings, self._test_targets_rs2v)

        self.assertAlmostEqual(r2c_dict["metrics"]["loss"], 1.8399004936218262, delta=self.error_tolerance,
                               msg="Loss not as expected for r2c!")
        self.assertAlmostEqual(r2v_dict["metrics"]["loss"], 66.42302047378135, delta=self.error_tolerance,
                               msg="Loss not as expected for r2v!")
        self.assertAlmostEqual(rs2c_dict["metrics"]["loss"], 1.7416267395019531, delta=self.error_tolerance,
                               msg="Loss not as expected for rs2c!")
        self.assertAlmostEqual(s2c_dict["metrics"]["loss"], 1.3706077337265015, delta=self.error_tolerance,
                               msg="Loss not as expected for s2c!")
        self.assertAlmostEqual(s2v_dict["metrics"]["loss"], 1.2276635438517511, delta=self.error_tolerance,
                               msg="Loss not as expected for s2v!")
        self.assertAlmostEqual(rs2v_dict["metrics"]["loss"], 107.96429166959688, delta=self.error_tolerance,
                               msg="Loss not as expected for rs2v!")

    def test_from_embeddings_with_bootstrapping(self):
        """
        Checks that metrics calculated with and without bootstrapping are within about the same range
        """

        def _get_error_range(bootstrapping_dict):
            return (bootstrapping_dict["upper"] - bootstrapping_dict["lower"]) * 0.5 * self.error_tolerance_factor

        # residue_to_class
        r2c_dict_bootstrapping = self.inferencer_r2c.from_embeddings_with_bootstrapping(self.per_residue_embeddings,
                                                                                        self._test_targets_r2c,
                                                                                        iterations=10,
                                                                                        sample_size=3,
                                                                                        seed=42)
        r2c_dict = self.inferencer_r2c.from_embeddings(self.per_residue_embeddings, self._test_targets_r2c)
        for metric in r2c_dict_bootstrapping.keys():
            self.assertAlmostEqual(r2c_dict["metrics"][metric], r2c_dict_bootstrapping[metric]["mean"],
                                   delta=_get_error_range(r2c_dict_bootstrapping[metric]))

        # residues_to_class
        rs2c_dict_bootstrapping = self.inferencer_rs2c.from_embeddings_with_bootstrapping(self.per_residue_embeddings,
                                                                                          self._test_targets_rs2c,
                                                                                          iterations=6,
                                                                                          sample_size=2,
                                                                                          seed=42)
        rs2c_dict = self.inferencer_rs2c.from_embeddings(self.per_residue_embeddings, self._test_targets_rs2c)
        for metric in rs2c_dict_bootstrapping.keys():
            self.assertAlmostEqual(rs2c_dict["metrics"][metric], rs2c_dict_bootstrapping[metric]["mean"],
                                   delta=_get_error_range(rs2c_dict_bootstrapping[metric]))

        # sequence_to_class
        s2c_dict_bootstrapping = self.inferencer_s2c.from_embeddings_with_bootstrapping(self.per_sequence_embeddings,
                                                                                        self._test_targets_s2c,
                                                                                        iterations=10,
                                                                                        sample_size=3,
                                                                                        seed=42)
        s2c_dict = self.inferencer_s2c.from_embeddings(self.per_sequence_embeddings, self._test_targets_s2c)
        for metric in s2c_dict_bootstrapping.keys():
            self.assertAlmostEqual(s2c_dict["metrics"][metric], s2c_dict_bootstrapping[metric]["mean"],
                                   delta=_get_error_range(s2c_dict_bootstrapping[metric]))

        # sequence_to_value
        s2v_dict_bootstrapping = self.inferencer_s2v.from_embeddings_with_bootstrapping(self.per_sequence_embeddings,
                                                                                        self._test_targets_s2v,
                                                                                        iterations=10,
                                                                                        sample_size=2,
                                                                                        seed=42)
        s2v_dict = self.inferencer_s2v.from_embeddings(self.per_sequence_embeddings, self._test_targets_s2v)
        for metric in s2v_dict_bootstrapping.keys():
            self.assertAlmostEqual(s2v_dict["metrics"][metric], s2v_dict_bootstrapping[metric]["mean"],
                                   delta=_get_error_range(s2v_dict_bootstrapping[metric]))

        # residues_to_value
        rs2v_dict_bootstrapping = self.inferencer_rs2v.from_embeddings_with_bootstrapping(self.per_residue_embeddings,
                                                                                          self._test_targets_rs2v,
                                                                                          iterations=10,
                                                                                          sample_size=2,
                                                                                          seed=42)
        rs2v_dict = self.inferencer_rs2v.from_embeddings(self.per_residue_embeddings, self._test_targets_rs2v)
        for metric in rs2v_dict_bootstrapping.keys():
            self.assertAlmostEqual(rs2v_dict["metrics"][metric], rs2v_dict_bootstrapping[metric]["mean"],
                                   delta=_get_error_range(rs2v_dict_bootstrapping[metric]))

        # Check that a sample size and iteration of 1 work
        _ = self.inferencer_r2c.from_embeddings_with_bootstrapping(self.per_residue_embeddings,
                                                                   self._test_targets_r2c,
                                                                   iterations=1,
                                                                   sample_size=3,
                                                                   seed=42)
        _ = self.inferencer_s2v.from_embeddings_with_bootstrapping(self.per_sequence_embeddings,
                                                                   self._test_targets_s2v,
                                                                   iterations=10,
                                                                   sample_size=1,
                                                                   seed=42)
        # Check that larger sample sizes than provided sequence number work
        _ = self.inferencer_s2v.from_embeddings_with_bootstrapping(self.per_sequence_embeddings,
                                                                   self._test_targets_s2v,
                                                                   iterations=5,
                                                                   sample_size=100,
                                                                   seed=42)

    def test_from_embeddings_with_dropout(self):
        # Test with different parameters
        r2c_dict = self.inferencer_r2c.from_embeddings_with_monte_carlo_dropout(
            self.per_residue_embeddings,
            n_forward_passes=10,
            confidence_level=0.05,
            seed=42
        )
        r2v_dict = self.inferencer_r2v.from_embeddings_with_monte_carlo_dropout(
            self.per_residue_embeddings,
            n_forward_passes=10,
            confidence_level=0.05,
            seed=42
        )
        rs2c_dict = self.inferencer_rs2c.from_embeddings_with_monte_carlo_dropout(
            self.per_residue_embeddings,
            n_forward_passes=15,
            confidence_level=0.5,
            seed=42
        )
        s2c_dict = self.inferencer_s2c.from_embeddings_with_monte_carlo_dropout(
            self.per_sequence_embeddings,
            n_forward_passes=2,
            confidence_level=0.01,
            seed=42
        )
        s2v_dict = self.inferencer_s2v.from_embeddings_with_monte_carlo_dropout(
            self.per_sequence_embeddings,
            n_forward_passes=10,
            confidence_level=0.05,
            seed=42
        )
        rs2v_dict = self.inferencer_rs2v.from_embeddings_with_monte_carlo_dropout(
            self.per_residue_embeddings,
            n_forward_passes=10,
            confidence_level=0.05,
            seed=42
        )

        # Basic structure tests
        self.assertTrue(all([len(r2c_dict[seq]) == len(self._test_sequences[idx])
                             for idx, seq in enumerate(r2c_dict.keys())]),
                        msg="Missing predictions for r2c!")

        self.assertTrue(all([len(r2v_dict[seq]) == len(self._test_sequences[idx])
                             for idx, seq in enumerate(r2v_dict.keys())]),
                        msg="Missing predictions for r2v!")

        # Check required keys exist
        required_keys = ["prediction", "all_predictions", "mcd_mean",
                         "mcd_lower_bound", "mcd_upper_bound"]

        for dict_to_check in [rs2c_dict, s2c_dict, s2v_dict, rs2v_dict]:
            for key in required_keys:
                self.assertTrue(key in list(dict_to_check.values())[0].keys(),
                                f"Missing {key} in predictions!")

        # Check confidence bounds
        for dict_to_check in [rs2c_dict, s2c_dict, s2v_dict, rs2v_dict]:
            for pred in dict_to_check.values():
                self.assertTrue(np.all((pred["mcd_lower_bound"] <= pred["mcd_mean"]).numpy()),
                                f"Lower bound greater than mean ({pred})!")
                self.assertTrue(np.all((pred["mcd_upper_bound"] >= pred["mcd_mean"]).numpy()),
                                f"Upper bound less than mean ({pred})!")

        # Check number of forward passes
        self.assertEqual(len(s2c_dict["Seq0"]["all_predictions"]), 2,
                         "Incorrect number of forward passes for s2c!")
        self.assertEqual(len(r2c_dict["Seq0"][0]["all_predictions"]), 10,
                         "Incorrect number of forward passes for r2c!")

        # Check prediction types
        self.assertTrue(all([type(s2v_dict[key]["prediction"]) is float for key in s2v_dict.keys()]),
                        msg="Regression prediction is not float!")
        self.assertTrue(all([type(rs2v_dict[key]["prediction"]) is float for key in rs2v_dict.keys()]),
                        msg="Regression prediction is not float!")

        # Test reproducibility with same seed
        s2v_dict_repeat = self.inferencer_s2v.from_embeddings_with_monte_carlo_dropout(
            self.per_sequence_embeddings,
            n_forward_passes=10,
            confidence_level=0.05,
            seed=42
        )
        for key in s2v_dict.keys():
            np.testing.assert_array_almost_equal(
                s2v_dict[key]["all_predictions"],
                s2v_dict_repeat[key]["all_predictions"],
                decimal=5,
                err_msg="MC Dropout not reproducible with same seed!"
            )

        # Test different seeds give different results
        s2v_dict_different_seed = self.inferencer_s2v.from_embeddings_with_monte_carlo_dropout(
            self.per_sequence_embeddings,
            n_forward_passes=10,
            confidence_level=0.05,
            seed=24
        )
        for key in s2v_dict.keys():
            self.assertFalse(
                np.array_equal(s2v_dict[key]["all_predictions"],
                               s2v_dict_different_seed[key]["all_predictions"]),
                "Different seeds produce identical results!"
            )

        # Test invalid parameters
        with self.assertRaises(Exception):
            self.inferencer_s2v.from_embeddings_with_monte_carlo_dropout(
                self.per_sequence_embeddings,
                n_forward_passes=0,  # Invalid number of passes
                confidence_level=0.05,
                seed=42
            )

        with self.assertRaises(Exception):
            self.inferencer_r2c.from_embeddings_with_monte_carlo_dropout(
                self.per_sequence_embeddings,
                n_forward_passes=1,  # Invalid number of passes
                confidence_level=0.05,
                seed=42
            )

        with self.assertRaises(Exception):
            self.inferencer_s2v.from_embeddings_with_monte_carlo_dropout(
                self.per_sequence_embeddings,
                n_forward_passes=10,
                confidence_level=1.5,  # Invalid confidence level
                seed=42
            )

    def _compare_predictions(self, preds: Dict[str, Union[float, List, List[List]]],
                             other_preds: Dict[str, Union[float, List, List[List]]]) -> List[str]:
        """
        Compare two predictions, handling both per-sequence and per-residue cases for classification
        and regression.

        :param preds: Dictionary of predictions
        :param other_preds: Dictionary of other predictions to compare
        :return: List of error messages, empty if no errors
        """
        error_messages = []
        for key in preds.keys():
            pred = np.array(preds[key])
            pred_other = np.array(other_preds[key])

            if pred.shape != pred_other.shape:
                error_messages.append(
                    f"Shape mismatch for key {key}: Predictions {pred.shape} vs other predictions {pred_other.shape}")
                continue

            diff = np.abs(pred - pred_other)
            max_diff = np.max(diff)

            if max_diff > self.error_tolerance:
                error_location = np.unravel_index(np.argmax(diff), diff.shape)
                error_messages.append(
                    f"Prediction mismatch for key {key} at index {error_location}: "
                    f"Predictions {pred[error_location]:.6f} vs other predictions {pred_other[error_location]:.6f}, "
                    f"difference {max_diff:.6f}"
                )

        return error_messages

    def test_onnx_conversion(self):
        return  # Don't run this test anymore, because of ONNX stability issues in the CI pipeline

        for inferencer in self.inferencer_list:
            with tempfile.TemporaryDirectory() as tmp_dir_name:
                # Convert
                converted_file_paths = inferencer.convert_to_onnx(tmp_dir_name)
                self.assertTrue(len(converted_file_paths) == 1)

                # Load converted model and make predictions
                model_path = converted_file_paths[0]
                embeddings = self.per_sequence_embeddings \
                    if inferencer.protocol in Protocol.using_per_sequence_embeddings() else self.per_residue_embeddings

                onnx_result_dict = Inferencer.from_onnx_with_embeddings(model_path=model_path,
                                                                        embeddings=embeddings,
                                                                        protocol=inferencer.protocol)

                # Compare to inferencer predictions
                inferencer_result_dict = inferencer.from_embeddings(embeddings=embeddings,
                                                                    include_probabilities=True)
                inferencer_result_dict_prob = inferencer_result_dict["mapped_probabilities"]

                prediction_errors = self._compare_predictions(preds=onnx_result_dict,
                                                              other_preds=inferencer_result_dict_prob)
                if len(prediction_errors) > 0:
                    print(prediction_errors)
                self.assertTrue(len(prediction_errors) == 0)

    def test_single_vs_batch_prediction(self):
        for inferencer in self.inferencer_list:
            embeddings = self.per_sequence_embeddings \
                if inferencer.protocol in Protocol.using_per_sequence_embeddings() else self.per_residue_embeddings

            pred_dict_batch = inferencer.from_embeddings(embeddings=embeddings, include_probabilities=True)

            pred_dict_single = {}
            for seq_id, emb in embeddings.items():
                pred_dict_single[seq_id] = inferencer.from_embeddings({seq_id: emb},
                                                                      include_probabilities=True)[
                    "mapped_probabilities"][seq_id]

            # Single predictions and batch predictions should not significantly differ
            prediction_errors = self._compare_predictions(pred_dict_batch["mapped_probabilities"], pred_dict_single)
            if len(prediction_errors) > 0:
                print(prediction_errors)
            self.assertTrue(len(prediction_errors) == 0)
